#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This module provides implementations of various colors and their respective colormodels,
    including: RGB, CMYK, grayscale, hex and X11
"""
import logging
import typing
from decimal import Decimal

from borb.io.read.types import Function, List, Name

logger = logging.getLogger(__name__)


class Color(object):
    """
    Color (American English), or colour (Commonwealth English),
    is the characteristic of visual perception described through color categories,
    with names such as red, orange, yellow, green, blue, or purple.
    This perception of color derives from the stimulation of photoreceptor cells
    (in particular cone cells in the human eye and other vertebrate eyes) by electromagnetic radiation
    (in the visible spectrum in the case of humans).
    Color categories and physical specifications of color are associated with objects through
    the wavelengths of the light that is reflected from them and their intensities.
    This reflection is governed by the object's physical properties such as light absorption, emission spectra, etc.
    """

    def to_rgb(self) -> "RGBColor":
        """
        This method returns the RGB representation of this Color
        """
        pass


class RGBColor(Color):
    """
    An RGB color space is any additive color space based on the RGB color model.
    A particular color space that employs RGB primaries for part of its specification is defined by the three chromaticities of the red, green, and blue additive primaries,
    and can produce any chromaticity that is the 2D triangle defined by those primary colors (ie. excluding transfer function, white point, etc.).
    The primary colors are specified in terms of their CIE 1931 color space chromaticity coordinates (x,y), linking them to human-visible color.
    RGB is an abbreviation for red–green–blue.
    """

    def __init__(self, r: Decimal, g: Decimal, b: Decimal):
        assert 0 <= r <= 1
        assert 0 <= g <= 1
        assert 0 <= b <= 1
        self.red = r
        self.green = g
        self.blue = b

    def to_rgb(self):
        """
        This method returns the RGB representation of this Color
        """
        return self

    def to_hex_string(self):
        """
        This method returns a hexadecimal string representing the RGB color
        """
        return "#{:02x}{:02x}{:02x}".format(
            int(self.red * 255), int(self.green * 255), int(self.blue * 255)
        )

    def __deepcopy__(self, memodict={}):
        return RGBColor(self.red, self.green, self.blue)


class CMYKColor(Color):
    """
    The CMYK color model (also known as process color, or four color) is a subtractive color model, based on the CMY color model,
    used in color printing, and is also used to describe the printing process itself.
    CMYK refers to the four ink plates used in some color printing: cyan, magenta, yellow, and key (black).

    The CMYK model works by partially or entirely masking colors on a lighter, usually white, background.
    The ink reduces the light that would otherwise be reflected.
    Such a model is called subtractive because inks "subtract" the colors red, green and blue from white light.
    White light minus red leaves cyan, white light minus green leaves magenta, and white light minus blue leaves yellow.

    In additive color models, such as RGB, white is the "additive" combination of all primary colored lights, black is the absence of light.
    In the CMYK model, it is the opposite: white is the natural color of the paper or other background,
    black results from a full combination of colored inks.
    To save cost on ink, and to produce deeper black tones, unsaturated and dark colors are produced by using black ink
    instead of the combination of cyan, magenta, and yellow.
    """

    def __init__(self, c: Decimal, m: Decimal, y: Decimal, k: Decimal):
        self.cyan = c
        self.magenta = m
        self.yellow = y
        self.key = k

    def to_rgb(self) -> "RGBColor":
        """
        This method returns the RGB representation of this Color
        """
        ONE = Decimal(1)
        r = (ONE - self.cyan) * (ONE - self.key)
        g = (ONE - self.magenta) * (ONE - self.key)
        b = (ONE - self.yellow) * (ONE - self.key)
        return RGBColor(r, g, b)

    def __deepcopy__(self, memodict={}):
        return CMYKColor(self.cyan, self.magenta, self.yellow, self.key)


class GrayColor(Color):
    """
    In digital photography, computer-generated imagery, and colorimetry,
    a grayscale or image is one in which the value of each pixel is a single sample representing only an amount of light;
    that is, it carries only intensity information.
    Grayscale images, a kind of black-and-white or gray monochrome, are composed exclusively of shades of gray.
    The contrast ranges from black at the weakest intensity to white at the strongest.
    """

    def __init__(self, g: Decimal):
        self.gray_level = g

    def to_rgb(self) -> "RGBColor":
        """
        This method returns the RGB representation of this Color
        """
        return RGBColor(self.gray_level, self.gray_level, self.gray_level)

    def __deepcopy__(self, memodict={}):
        return GrayColor(self.gray_level)


class HexColor(RGBColor):
    """
    A hex triplet is a six-digit, three-byte hexadecimal number used in HTML, CSS, SVG, and other computing applications to represent colors.
    The bytes represent the red, green, and blue components of the color.
    One byte represents a number in the range 00 to FF (in hexadecimal notation), or 0 to 255 in decimal notation.
    This represents the least (0) to the most (255) intensity of each of the color components.

    Thus web colors specify colors in the 24-bit RGB color scheme.

    The hex triplet is formed by concatenating three bytes in hexadecimal notation, in the following order:
    - Byte 1: red value (color type red)
    - Byte 2: green value (color type green)
    - Byte 3: blue value (color type blue)
    """

    def __init__(self, hex_string: str):
        if hex_string.startswith("#"):
            hex_string = hex_string[1:]
        assert len(hex_string) == 6 or len(hex_string) == 8
        r: float = 0
        g: float = 0
        b: float = 0
        a: float = 0
        if len(hex_string) == 6:
            a = 255
            r = int(hex_string[0:2], 16)
            g = int(hex_string[2:4], 16)
            b = int(hex_string[4:6], 16)
        if len(hex_string) == 8:
            a = int(hex_string[0:2], 16)
            r = int(hex_string[2:4], 16)
            g = int(hex_string[4:6], 16)
            b = int(hex_string[6:8], 16)
        a /= 255
        r /= 255
        g /= 255
        b /= 255
        super(HexColor, self).__init__(Decimal(r), Decimal(g), Decimal(b))

    def __deepcopy__(self, memodict={}):
        return HexColor(self.to_hex_string())


class HSVColor(Color):
    """
    HSL (hue, saturation, lightness) and HSV (hue, saturation, value, also known as HSB or hue, saturation, brightness)
    are alternative representations of the RGB color model, designed in the 1970s by computer graphics researchers
    to more closely align with the way human vision perceives color-making attributes.
    In these models, colors of each hue are arranged in a radial slice,
    around a central axis of neutral colors which ranges from black at the bottom to white at the top.

    The HSL representation models the way different paints mix together to create colour in the real world,
    with the lightness dimension resembling the varying amounts of black or white paint in the mixture
    (e.g. to create "light red", a red pigment can be mixed with white paint;
    this white paint corresponds to a high "lightness" value in the HSL representation).
    Fully saturated colors are placed around a circle at a lightness value of ½,
    with a lightness value of 0 or 1 corresponding to fully black or white, respectively.

    Meanwhile, the HSV representation models how colors appear under light.
    The difference between HSL and HSV is that a color with maximum lightness in HSL is pure white,
    but a color with maximum value/brightness in HSV is analogous to shining a white light on a colored object
    (e.g. shining a bright white light on a red object causes the object to still appear red, just brighter and more intense,
    while shining a dim light on a red object causes the object to appear darker and less bright).
    """

    def __init__(self, hue: Decimal, saturation: Decimal, value: Decimal):
        self.hue = hue
        self.saturation = saturation
        self.value = value

    def to_rgb(self) -> "RGBColor":
        """
        This method returns the RGB representation of this Color
        """
        h, s, v = self.hue, self.saturation, self.value
        ONE = Decimal(1)
        SIX = Decimal(6)
        if s == 0:
            return RGBColor(Decimal(v), Decimal(v), Decimal(v))
        i = int(h * SIX)  # XXX assume int() truncates!
        f = (h * SIX) - i
        p, q, t = v * (ONE - s), v * (ONE - s * f), v * (ONE - s * (ONE - f))
        i %= 6
        if i == 0:
            return RGBColor(Decimal(v), Decimal(t), Decimal(p))
        if i == 1:
            return RGBColor(Decimal(q), Decimal(v), Decimal(p))
        if i == 2:
            return RGBColor(Decimal(p), Decimal(v), Decimal(t))
        if i == 3:
            return RGBColor(Decimal(p), Decimal(q), Decimal(v))
        if i == 4:
            return RGBColor(Decimal(t), Decimal(p), Decimal(v))
        if i == 5:
            return RGBColor(Decimal(v), Decimal(p), Decimal(q))
        return RGBColor(Decimal(0), Decimal(0), Decimal(0))

    @staticmethod
    def from_rgb(c: RGBColor) -> "HSVColor":
        """
        This method returns the HSV representation of an RGB color
        """
        RGB_MAX = Decimal(255)
        r, g, b = c.red, c.green, c.blue
        mx = max(r, g, b)
        mn = min(r, g, b)
        df = mx - mn
        if mx == mn:
            h = Decimal(0)
        elif mx == r:
            h = (Decimal(60) * ((g - b) / df) + Decimal(360)) % Decimal(360)
        elif mx == g:
            h = (Decimal(60) * ((b - r) / df) + Decimal(120)) % Decimal(360)
        elif mx == b:
            h = (Decimal(60) * ((r - g) / df) + Decimal(240)) % Decimal(360)
        if mx == 0:
            s = Decimal(0)
        else:
            s = (df / mx) * 100
        v = mx * 100
        HUNDRED = Decimal(100)
        FULL_CIRCLE = Decimal(360)
        return HSVColor(h / FULL_CIRCLE, s / HUNDRED, v / HUNDRED)

    def darker(self) -> "HSVColor":
        """
        This function returns a darker shade of the current HSV color
        """
        return HSVColor(self.hue, self.saturation, self.value * Decimal(0.8))

    @staticmethod
    def complementary(color: Color):
        """
        This function returns an HSV color whose hue is the complement of the current HSV color
        """
        c: HSVColor = HSVColor.from_rgb(color.to_rgb())
        new_hue: int = int(float(c.hue) * 360.0) + 180 % 360
        return HSVColor(Decimal(new_hue / 360), c.saturation, c.value)

    @staticmethod
    def analogous(color: Color) -> typing.List[Color]:
        """
        This function returns an analogous color scheme.
        Analogous color schemes use colors that are next to each other on the color wheel. They usually match well and create serene and comfortable designs.
        Analogous color schemes are often found in nature and are harmonious and pleasing to the eye.
        """
        c: HSVColor = HSVColor.from_rgb(color.to_rgb())
        return [
            HSVColor(Decimal((int(c.hue * 360 + a) % 360) / 360), c.saturation, c.value)
            for a in [0, 15, 30]
        ]

    @staticmethod
    def split_complementary(color: Color) -> typing.List[Color]:
        """
        This function returns a split complementary color scheme.
        A split complementary color scheme uses two colors plus the color that is opposite to them on the color wheel.
        For example blue and purple with yellow.
        :param color:       the complementary color
        :return:            3 colors (a split complementary color scheme)
        """
        c: HSVColor = HSVColor.from_rgb(color.to_rgb())
        return [
            HSVColor(Decimal((int(c.hue * 360 + a) % 360) / 360), c.saturation, c.value)
            for a in [0, 165, 195]
        ]

    @staticmethod
    def triadic(color: Color) -> typing.List[Color]:
        """
        This function returns a triadic color scheme.
        A triadic color scheme uses colors that are evenly spaced around the color wheel.
        Triadic color harmonies tend to be quite vibrant, even if you use pale or unsaturated versions of your hues.
        To use a triadic harmony successfully, the colors should be carefully balanced - let one color dominate and use the two others for accent.
        """
        c: HSVColor = HSVColor.from_rgb(color.to_rgb())
        return [
            HSVColor(Decimal((int(c.hue * 360 + a) % 360) / 360), c.saturation, c.value)
            for a in [0, 60, 120]
        ]

    @staticmethod
    def tetradic_rectangle(c: Color) -> typing.List[Color]:
        """
        This function returns a tetradic (rectangular) color scheme.
        The rectangle or tetradic color scheme uses four colors arranged into two complementary pairs.
        This rich color scheme offers plenty of possibilities for variation.
        Tetradic color schemes works best if you let one color be dominant.
        :param c:   one color of one of the complementary paris
        :return:    four colors (two complementary pairs)
        """
        c0: HSVColor = HSVColor.from_rgb(c.to_rgb())
        return [
            HSVColor(
                Decimal((int(c0.hue * 360 + a) % 360) / 360), c0.saturation, c0.value
            )
            for a in [0, 30, 180, 210]
        ]

    @staticmethod
    def tetradic_square(color: Color) -> typing.List[Color]:
        """
        This function returns a tetradic (square) color scheme.
        This harmony is comparable to the Tetradic harmony but with the four colors spaced evenly around the color wheel.
        This harmony works best if one color dominates.
        Just like the Tetradic harmony, you need to keep track of the relationship between the cool and the warm colors.
        :param color:
        :return:        four colors (a tetradic square color scheme)
        """
        c: HSVColor = HSVColor.from_rgb(color.to_rgb())
        return [
            HSVColor(Decimal((int(c.hue * 360 + a) % 360) / 360), c.saturation, c.value)
            for a in [0, 90, 180, 270]
        ]

    def __deepcopy__(self, memodict={}):
        return HSVColor(self.hue, self.saturation, self.value)


class X11Color(HexColor):
    """
    In computing, on the X Window System, X11 color names are represented in a simple text file,
    which maps certain strings to RGB color values.
    It was traditionally shipped with every X11 installation,
    hence the name, and is usually located in <X11root>/lib/X11/rgb.txt.
    The web colors list is descended from it but differs for certain color names.
    """

    COLOR_DEFINITION = {
        "AliceBlue": "#FFF0F8FF",
        "AntiqueWhite": "#FFFAEBD7",
        "Aqua": "#FF00FFFF",
        "Aquamarine": "#FF7FFFD4",
        "Azure": "#FFF0FFFF",
        "Beige": "#FFF5F5DC",
        "Bisque": "#FFFFE4C4",
        "Black": "#FF000000",
        "BlanchedAlmond": "#FFFFEBCD",
        "Blue": "#FF0000FF",
        "BlueViolet": "#FF8A2BE2",
        "Brown": "#FFA52A2A",
        "BurlyWood": "#FFDEB887",
        "CadetBlue": "#FF5F9EA0",
        "Chartreuse": "#FF7FFF00",
        "Chocolate": "#FFD2691E",
        "Coral": "#FFFF7F50",
        "CornflowerBlue": "#FF6495ED",
        "Cornsilk": "#FFFFF8DC",
        "Crimson": "#FFDC143C",
        "Cyan": "#FF00FFFF",
        "DarkBlue": "#FF00008B",
        "DarkCyan": "#FF008B8B",
        "DarkGoldenrod": "#FFB8860B",
        "DarkGray": "#FFA9A9A9",
        "DarkGreen": "#FF006400",
        "DarkKhaki": "#FFBDB76B",
        "DarkMagenta": "#FF8B008B",
        "DarkOliveGreen": "#FF556B2F",
        "DarkOrange": "#FFFF8C00",
        "DarkOrchid": "#FF9932CC",
        "DarkRed": "#FF8B0000",
        "DarkSalmon": "#FFE9967A",
        "DarkSeaGreen": "#FF8FBC8F",
        "DarkSlateBlue": "#FF483D8B",
        "DarkSlateGray": "#FF2F4F4F",
        "DarkTurquoise": "#FF00CED1",
        "DarkViolet": "#FF9400D3",
        "DeepPink": "#FFFF1493",
        "DeepSkyBlue": "#FF00BFFF",
        "DimGray": "#FF696969",
        "DodgerBlue": "#FF1E90FF",
        "Firebrick": "#FFB22222",
        "FloralWhite": "#FFFFFAF0",
        "ForestGreen": "#FF228B22",
        "Fuchsia": "#FFFF00FF",
        "Gainsboro": "#FFDCDCDC",
        "GhostWhite": "#FFF8F8FF",
        "Gold": "#FFFFD700",
        "Goldenrod": "#FFDAA520",
        "Gray": "#FF808080",
        "Green": "#FF008000",
        "GreenYellow": "#FFADFF2F",
        "Honeydew": "#FFF0FFF0",
        "HotPink": "#FFFF69B4",
        "IndianRed": "#FFCD5C5C",
        "Indigo": "#FF4B0082",
        "Ivory": "#FFFFFFF0",
        "Khaki": "#FFF0E68C",
        "Lavender": "#FFE6E6FA",
        "LavenderBlush": "#FFFFF0F5",
        "LawnGreen": "#FF7CFC00",
        "LemonChiffon": "#FFFFFACD",
        "LightBlue": "#FFADD8E6",
        "LightCoral": "#FFF08080",
        "LightCyan": "#FFE0FFFF",
        "LightGoldenrodYellow": "#FFFAFAD2",
        "LightGray": "#FFD3D3D3",
        "LightGreen": "#FF90EE90",
        "LightPink": "#FFFFB6C1",
        "LightSalmon": "#FFFFA07A",
        "LightSeaGreen": "#FF20B2AA",
        "LightSkyBlue": "#FF87CEFA",
        "LightSlateGray": "#FF778899",
        "LightSteelBlue": "#FFB0C4DE",
        "LightYellow": "#FFFFFFE0",
        "Lime": "#FF00FF00",
        "LimeGreen": "#FF32CD32",
        "Linen": "#FFFAF0E6",
        "Magenta": "#FFFF00FF",
        "Maroon": "#FF800000",
        "MediumAquamarine": "#FF66CDAA",
        "MediumBlue": "#FF0000CD",
        "MediumOrchid": "#FFBA55D3",
        "MediumPurple": "#FF9370DB",
        "MediumSeaGreen": "#FF3CB371",
        "MediumSlateBlue": "#FF7B68EE",
        "MediumSpringGreen": "#FF00FA9A",
        "MediumTurquoise": "#FF48D1CC",
        "MediumVioletRed": "#FFC71585",
        "MidnightBlue": "#FF191970",
        "MintCream": "#FFF5FFFA",
        "MistyRose": "#FFFFE4E1",
        "Moccasin": "#FFFFE4B5",
        "NavajoWhite": "#FFFFDEAD",
        "Navy": "#FF000080",
        "OldLace": "#FFFDF5E6",
        "Olive": "#FF808000",
        "OliveDrab": "#FF6B8E23",
        "Orange": "#FFFFA500",
        "OrangeRed": "#FFFF4500",
        "Orchid": "#FFDA70D6",
        "PaleGoldenrod": "#FFEEE8AA",
        "PaleGreen": "#FF98FB98",
        "PaleTurquoise": "#FFAFEEEE",
        "PaleVioletRed": "#FFDB7093",
        "PapayaWhip": "#FFFFEFD5",
        "PeachPuff": "#FFFFDAB9",
        "Peru": "#FFCD853F",
        "Pink": "#FFFFC0CB",
        "Plum": "#FFDDA0DD",
        "PowderBlue": "#FFB0E0E6",
        "Purple": "#FF800080",
        "Red": "#FFFF0000",
        "RosyBrown": "#FFBC8F8F",
        "RoyalBlue": "#FF4169E1",
        "SaddleBrown": "#FF8B4513",
        "Salmon": "#FFFA8072",
        "SandyBrown": "#FFF4A460",
        "SeaGreen": "#FF2E8B57",
        "SeaShell": "#FFFFF5EE",
        "Sienna": "#FFA0522D",
        "Silver": "#FFC0C0C0",
        "SkyBlue": "#FF87CEEB",
        "SlateBlue": "#FF6A5ACD",
        "SlateGray": "#FF708090",
        "Snow": "#FFFFFAFA",
        "SpringGreen": "#FF00FF7F",
        "SteelBlue": "#FF4682B4",
        "Tan": "#FFD2B48C",
        "Teal": "#FF008080",
        "Thistle": "#FFD8BFD8",
        "Tomato": "#FFFF6347",
        "Transparent": "#00FFFFFF",
        "Turquoise": "#FF40E0D0",
        "Violet": "#FFEE82EE",
        "Wheat": "#FFF5DEB3",
        "White": "#FFFFFFFF",
        "WhiteSmoke": "#FFF5F5F5",
        "Yellow": "#FFFFFF00",
        "YellowGreen": "#FF9ACD32",
    }

    def __init__(self, color_name: str):
        assert color_name in X11Color.COLOR_DEFINITION
        self.color_name: str = color_name
        super(X11Color, self).__init__(X11Color.COLOR_DEFINITION[color_name])

    def get_name(self) -> str:
        """
        This function returns the name of this X11Color
        """
        return self.color_name

    @staticmethod
    def find_nearest_x11_color(color: Color) -> "X11Color":
        """
        This function find the nearest X11Color equivalent for a given Color
        """
        rgb_color_001: RGBColor = color.to_rgb()
        d_min: typing.Optional[Decimal] = None
        c_min: typing.Optional[str] = None
        for n, c in X11Color.COLOR_DEFINITION.items():
            rgb_color_002: RGBColor = HexColor(c)
            d: Decimal = (
                (rgb_color_001.red - rgb_color_002.red) ** 2
                + (rgb_color_001.green - rgb_color_002.green) ** 2
                + (rgb_color_001.blue - rgb_color_002.blue) ** 2
            )
            if d_min is None or d < d_min:
                d_min = d
                c_min = n
        assert d_min is not None
        assert c_min is not None
        return X11Color(c_min)

    def __deepcopy__(self, memodict={}):
        return X11Color(self.color_name)


class Separation(Color):
    """
    When printing a page, most devices produce a single composite page on which all process colorants (and spot
    colorants, if any) are combined. However, some devices, such as imagesetters, produce a separate,
    monochromatic rendition of the page, called a separation, for each colorant. When the separations are later
    combined—on a printing press, for example—and the proper inks or other colorants are applied to them, the
    result is a full-colour page.
    """

    def __init__(self, color_space: typing.List[typing.Any], xs: typing.List[Decimal]):
        super(Separation, self).__init__()
        self.color_space = color_space
        self.xs = xs
        self._to_rgb_cache: typing.Optional[RGBColor] = None
        """"
        except Exception as ex:
            if len(color_space) != 0 and len(xs) != 0:
                logger.debug("Unable to instantiate separation color, defaulting to black")
            self.rgb_color = RGBColor(Decimal(0), Decimal(0), Decimal(0))
            pass
        """

    def to_rgb(self) -> "RGBColor":
        """
        This method returns the RGB representation of this Color
        """
        if self._to_rgb_cache is not None:
            return self._to_rgb_cache

        assert isinstance(self.color_space[3], Function)
        tint_function: Function = self.color_space[3]

        # run function
        try:
            ys: typing.List[Decimal] = tint_function.evaluate(self.xs)
            assert ys is not None
        except:
            logger.info(
                "Unable to execute TintFunction for Separation %d, defaulting to black"
                % id(self)
            )
            self._to_rgb_cache = RGBColor(Decimal(0), Decimal(0), Decimal(0))
            return self._to_rgb_cache

        # determine the color space to map to
        alternative_color_space: typing.Optional[Name] = None
        if isinstance(self.color_space[2], Name):
            alternative_color_space = self.color_space[2]
        if (
            isinstance(self.color_space[2], List)
            and len(self.color_space[2]) == 2
            and isinstance(self.color_space[2][0], Name)
        ):
            alternative_color_space = self.color_space[2][0]

        # DeviceCMYK
        if alternative_color_space == "DeviceCMYK":
            self._to_rgb_cache = CMYKColor(ys[0], ys[1], ys[2], ys[3]).to_rgb()
            return self._to_rgb_cache

        # DeviceRGB
        if alternative_color_space == "DeviceRGB":
            self._to_rgb_cache = RGBColor(ys[0], ys[1], ys[2])
            return self._to_rgb_cache

        # ICCBased
        if alternative_color_space == "ICCBased":
            if len(ys) == 1:
                self._to_rgb_cache = GrayColor(ys[0]).to_rgb()
            if len(ys) == 3:
                self._to_rgb_cache = RGBColor(ys[0], ys[1], ys[2])
            if len(ys) == 4:
                self._to_rgb_cache = CMYKColor(ys[0], ys[1], ys[2], ys[3]).to_rgb()
            assert self._to_rgb_cache is not None
            return self._to_rgb_cache

        # default
        return RGBColor(Decimal(0), Decimal(0), Decimal(0))

    def __deepcopy__(self, memodict={}):
        out: Separation = Separation([], [])
        out._to_rgb_cache = self.to_rgb()
        return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This class represents a Table with columns that will assume
a width based on their contents. It tries to emulate the behaviour
of <table> elements in HTML
"""
import math
import typing
from decimal import Decimal

from borb.pdf.canvas.color.color import Color, HexColor
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.layout_element import Alignment
from borb.pdf.canvas.layout.table.table import Table
from borb.pdf.canvas.layout.text.paragraph import Paragraph
from borb.pdf.page.page import Page


class FlexibleColumnWidthTable(Table):
    """
    This class represents a Table with columns that will assume
    a width based on their contents. It tries to emulate the behaviour
    of <table> elements in HTML
    """

    def __init__(
        self,
        number_of_rows: int,
        number_of_columns: int,
        border_top: bool = False,
        border_right: bool = False,
        border_bottom: bool = False,
        border_left: bool = False,
        border_radius_bottom_left: Decimal = Decimal(0),
        border_radius_bottom_right: Decimal = Decimal(0),
        border_radius_top_left: Decimal = Decimal(0),
        border_radius_top_right: Decimal = Decimal(0),
        border_color: Color = HexColor("000000"),
        border_width: Decimal = Decimal(1),
        padding_top: Decimal = Decimal(0),
        padding_right: Decimal = Decimal(0),
        padding_bottom: Decimal = Decimal(0),
        padding_left: Decimal = Decimal(0),
        margin_top: Decimal = Decimal(0),
        margin_right: Decimal = Decimal(0),
        margin_bottom: Decimal = Decimal(0),
        margin_left: Decimal = Decimal(0),
        horizontal_alignment: Alignment = Alignment.LEFT,
        vertical_alignment: Alignment = Alignment.TOP,
        background_color: typing.Optional[Color] = None,
    ):
        super(FlexibleColumnWidthTable, self).__init__(
            number_of_rows=number_of_rows,
            number_of_columns=number_of_columns,
            background_color=background_color,
            border_bottom=border_bottom,
            border_color=border_color,
            border_left=border_left,
            border_radius_bottom_left=border_radius_bottom_left,
            border_radius_bottom_right=border_radius_bottom_right,
            border_radius_top_left=border_radius_top_left,
            border_radius_top_right=border_radius_top_right,
            border_right=border_right,
            border_top=border_top,
            border_width=border_width,
            horizontal_alignment=horizontal_alignment,
            margin_bottom=margin_bottom if margin_bottom is not None else Decimal(5),
            margin_left=margin_left if margin_left is not None else Decimal(5),
            margin_right=margin_right if margin_right is not None else Decimal(5),
            margin_top=margin_top if margin_top is not None else Decimal(5),
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            padding_right=padding_right,
            padding_top=padding_top,
            vertical_alignment=vertical_alignment,
        )

    def _get_min_column_width(self, col: int) -> Decimal:
        widths: typing.List[Decimal] = []
        for table_cell in [
            x for x in self._get_cells_at_column(col) if x._col_span == 1
        ]:
            if table_cell._min_width is None:
                widths.append(Decimal(0))
                continue
            if table_cell._preferred_width is None:
                assert table_cell._min_width is not None
                widths.append(table_cell._min_width)
                continue
            if table_cell._preferred_width > table_cell._min_width:
                assert table_cell._preferred_width is not None
                widths.append(table_cell._preferred_width)
                continue
            # default
            assert table_cell._min_width is not None
            widths.append(table_cell._min_width)

        # exception
        if len(widths) == 0:
            return Decimal(0)

        # return
        return max(widths)

    def _get_max_column_width(self, col: int) -> Decimal:
        widths: typing.List[Decimal] = []
        for table_cell in [
            x for x in self._get_cells_at_column(col) if x._col_span == 1
        ]:
            if table_cell._max_width is None:
                widths.append(Decimal(2048))
                continue
            if table_cell._preferred_width is None:
                widths.append(table_cell._max_width)
                continue
            if table_cell._preferred_width < table_cell._max_width:
                widths.append(table_cell._preferred_width)
                continue
            # default
            widths.append(table_cell._max_width)

        # exception
        if len(widths) == 0:
            return Decimal(2048)

        # return
        return max(widths)

    def _get_grid_coordinates(
        self,
        available_space: Rectangle,  # type: ignore[name-defined]
    ) -> typing.List[typing.List[typing.Tuple[Decimal, Decimal]]]:
        # 1.    Calculate the minimum content width (MCW) of each cell: the formatted content may span any number of lines but may not overflow the cell box.
        #       If the specified 'width' (W) of the cell is greater than MCW, W is the minimum cell width.
        #       A value of 'auto' means that MCW is the minimum cell width.
        #
        #       Also, calculate the "maximum" cell width of each cell:
        #       formatting the content without breaking lines other than where explicit line breaks occur.
        for t in self._content:
            t._calculate_min_and_max_layout_box()

        # 2.    For each column, determine a maximum and minimum column width from the cells that span only that column.
        #       The minimum is that required by the cell with the largest minimum cell width (or the column 'width', whichever is larger).
        #       The maximum is that required by the cell with the largest maximum cell width (or the column 'width', whichever is larger).
        min_column_widths: typing.List[Decimal] = [
            self._get_min_column_width(i) for i in range(0, self._number_of_columns)
        ]
        max_column_widths: typing.List[Decimal] = [
            self._get_max_column_width(i) for i in range(0, self._number_of_columns)
        ]

        # 3.    For each cell that spans more than one column, increase the minimum widths of the columns it spans so that together,
        #       they are at least as wide as the cell. Do the same for the maximum widths.
        #       If possible, widen all spanned columns by approximately the same amount.
        for table_cell in self._content:
            if table_cell._col_span == 1:
                continue
            column_indices: typing.Set[int] = set(
                [y for x, y in table_cell._table_coordinates]
            )
            sum_of_min_col_spans: Decimal = sum(
                [min_column_widths[x] for x in column_indices]
            )
            if sum_of_min_col_spans < table_cell._min_width:
                delta: Decimal = table_cell._min_width - sum_of_min_col_spans
                min_column_widths = [
                    w + (delta / table_cell._col_span) if i in column_indices else w
                    for i, w in enumerate(min_column_widths)
                ]

            sum_of_max_col_spans: Decimal = sum(
                [max_column_widths[x] for x in column_indices]
            )
            if sum_of_max_col_spans < table_cell._max_width:
                delta: Decimal = table_cell._max_width - sum_of_max_col_spans
                max_column_widths = [
                    w + (delta / table_cell._col_span) if i in column_indices else w
                    for i, w in enumerate(max_column_widths)
                ]

        # 4.    For each column group element with a 'width' other than 'auto', increase the minimum widths of the columns it spans,
        #       so that together they are at least as wide as the column group's 'width'.
        #       This gives a maximum and minimum width for each column.

        # 5. calculate column width based on min, max and bounding box
        # start by assigning each column its minimum width
        column_widths: typing.List[Decimal] = [x for x in min_column_widths]

        # as long as there are columns that could be expanded, and there is room to expand them
        # add Decimal(1) to each (expandable) column_width
        number_of_expandable_columns: int = sum(
            [
                1
                for i in range(0, len(column_widths))
                if column_widths[i] < max_column_widths[i]
            ]
        )
        while (
            sum(column_widths) + number_of_expandable_columns
            < available_space.get_width()
            and number_of_expandable_columns > 0
        ):
            for i in range(0, len(column_widths)):
                if column_widths[i] < max_column_widths[i]:
                    column_widths[i] += Decimal(1)
            number_of_expandable_columns = sum(
                [
                    1
                    for i in range(0, len(column_widths))
                    if column_widths[i] < max_column_widths[i]
                ]
            )

        # convert grid coordinates to page coordinates
        grid_x_to_page_x: typing.List[Decimal] = [available_space.get_x()]
        for i in range(1, self._number_of_columns + 1):
            prev_x: Decimal = grid_x_to_page_x[-1]
            new_x: Decimal = prev_x + column_widths[i - 1]
            grid_x_to_page_x.append(new_x)

        grid_y_to_page_y: typing.List[Decimal] = [
            available_space.get_y() + available_space.get_height()
        ]

        # calculate bounds of TableCells with row_span == 1
        for r in range(0, self._number_of_rows):
            prev_row_lboxes: typing.List[Rectangle] = []
            for e in [x for x in self._get_cells_at_row(r) if x._row_span == 1]:

                # get coordinates of lower-left corner of this TableCell (in grid space)
                # table keeps track of things in (row, column) style
                # hence p[1], rather than p[0]
                grid_x: int = min([p[1] for p in e._table_coordinates])

                # layout
                h: Decimal = grid_y_to_page_y[r] - available_space.get_y()
                prev_row_lboxes.append(
                    e.get_layout_box(
                        Rectangle(
                            grid_x_to_page_x[grid_x],
                            available_space.get_y(),
                            grid_x_to_page_x[grid_x + e._col_span]
                            - grid_x_to_page_x[grid_x],
                            h,
                        )
                    )
                )

            # keep track of the bottom of the previous (at this point current) row
            # this makes it easier to lay out the next row
            new_y: Decimal = min([lbox.get_y() for lbox in prev_row_lboxes])
            grid_y_to_page_y.append(new_y)

        # return
        return [[(x, y) for y in grid_y_to_page_y] for x in grid_x_to_page_x]

    def _get_content_box(self, available_space: Rectangle) -> Rectangle:

        # fill table
        number_of_cells: int = self._number_of_rows * self._number_of_columns
        empty_cells: int = number_of_cells - sum(
            [(x._row_span * x._col_span) for x in self._content]
        )
        for _ in range(0, empty_cells):
            self.add(Paragraph(" ", respect_spaces_in_text=True))

        # return
        m = self._get_grid_coordinates(available_space)
        min_x: Decimal = m[0][0][0]
        max_x: Decimal = m[-1][-1][0]
        min_y: Decimal = m[-1][-1][1]
        max_y: Decimal = m[0][0][1]
        return Rectangle(
            available_space.get_x(),
            min_y,
            Decimal(math.ceil(max_x - min_x)),
            max_y - min_y,
        )

    def _paint_content_box(self, page: Page, available_space: Rectangle) -> None:

        # fill table
        number_of_cells: int = self._number_of_rows * self._number_of_columns
        empty_cells: int = number_of_cells - sum(
            [(x._row_span * x._col_span) for x in self._content]
        )
        for _ in range(0, empty_cells):
            self.add(Paragraph(" ", respect_spaces_in_text=True))

        m: typing.List[
            typing.List[typing.Tuple[Decimal, Decimal]]
        ] = self._get_grid_coordinates(available_space)

        # paint
        for e in self._content:
            grid_x: int = min([p[1] for p in e._table_coordinates])
            grid_y: int = min([p[0] for p in e._table_coordinates])
            x: Decimal = m[grid_x][grid_y][0]
            y: Decimal = m[grid_x][grid_y + e._row_span][1]
            w: Decimal = m[grid_x + e._col_span][grid_y][0] - x
            h: Decimal = m[grid_x][grid_y][1] - y
            cbox: Rectangle = Rectangle(x, y, w, h)
            e._set_layout_box(cbox)
            e.paint(page, cbox)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Set the text font, T f , to font and the text font size, T fs , to size. font shall be
the name of a font resource in the Font subdictionary of the current
resource dictionary; size shall be a number representing a scale factor.
There is no initial value for either font or size; they shall be specified
explicitly by using Tf before any text is shown.
"""

import typing

from borb.io.read.types import AnyPDFType, Decimal
from borb.pdf.canvas.font.font import Font
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class SetFontAndSize(CanvasOperator):
    """
    Set the text font, T f , to font and the text font size, T fs , to size. font shall be
    the name of a font resource in the Font subdictionary of the current
    resource dictionary; size shall be a number representing a scale factor.
    There is no initial value for either font or size; they shall be specified
    explicitly by using Tf before any text is shown.
    """

    def __init__(self):
        super().__init__("Tf", 2)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the Tf operator
        """

        # lookup font dictionary
        font_ref = canvas_stream_processor.get_resource("Font", operands[0])
        assert font_ref is not None
        assert isinstance(font_ref, Font)

        # font size
        font_size = operands[1]
        assert isinstance(font_size, Decimal)

        # set state
        canvas = canvas_stream_processor.get_canvas()
        canvas.graphics_state.font_size = font_size
        canvas.graphics_state.font = operands[
            0
        ]  # in stead of setting the Font in the graphics_state, we explictly set the Name of the Font. This is a lot cheaper to copy by the Q/q operator.



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of EventListener extracts all Image objects on a Page
"""
from typing import List

from PIL import Image  # type: ignore [import]

from borb.pdf.canvas.event.begin_page_event import BeginPageEvent
from borb.pdf.canvas.event.event_listener import Event, EventListener
from borb.pdf.canvas.event.image_render_event import ImageRenderEvent
from borb.pdf.page.page import Page


class SimpleImageExtraction(EventListener):
    """
    This implementation of EventListener extracts all Image objects on a Page
    """

    def __init__(self):
        """
        Constructs a new SimpleImageExtraction
        """
        self._image_render_info_per_page = {}
        self._current_page: int = -1

    def _event_occurred(self, event: "Event") -> None:
        if isinstance(event, BeginPageEvent):
            self._begin_page(event.get_page())
        if isinstance(event, ImageRenderEvent):
            self._render_image(event)

    def get_images_for_page(self, page_nr: int) -> List[Image.Image]:
        """
        This function returns a typing.List[Image] on a given page
        """
        return (
            self._image_render_info_per_page[page_nr]
            if page_nr in self._image_render_info_per_page
            else []
        )

    def _render_image(self, image_render_event: "ImageRenderEvent"):

        # init if needed
        if self._current_page not in self._image_render_info_per_page:
            self._image_render_info_per_page[self._current_page] = []

        # append ImageRenderEvent
        self._image_render_info_per_page[self._current_page].append(
            image_render_event.get_image()
        )

    def _begin_page(self, page: Page):
        self._current_page += 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
In turn, each PDF/A standard supports different conformance levels (a & b for PDF/A-1; and a, b & u for PDF/A-2 and -3).
These conformance levels control the “accessibility” requirements of a file that impact the ability of machines and people to understand the content.
"""

import enum


class ConformanceLevel(enum.Enum):
    """
    In turn, each PDF/A standard supports different conformance levels (a & b for PDF/A-1; and a, b & u for PDF/A-2 and -3).
    These conformance levels control the “accessibility” requirements of a file that impact the ability of machines and people to understand the content.
    """

    PDFA_1A = 2
    PDFA_1B = 3
    PDFA_2A = 5
    PDFA_2B = 7
    PDFA_2U = 11
    PDFA_3A = 13
    PDFA_3B = 17
    PDFA_3U = 19

    def get_standard(self) -> int:
        """
        This function returns the standard (the numeric part) of this ConformanceLevel
        :return:    the numeric part of this ConformanceLevel (1, 2, 3)
        """
        if self in [ConformanceLevel.PDFA_1A, ConformanceLevel.PDFA_1B]:
            return 1
        if self in [
            ConformanceLevel.PDFA_2A,
            ConformanceLevel.PDFA_2B,
            ConformanceLevel.PDFA_2U,
        ]:
            return 2
        if self in [
            ConformanceLevel.PDFA_3A,
            ConformanceLevel.PDFA_3B,
            ConformanceLevel.PDFA_3U,
        ]:
            return 3
        assert False

    def get_conformance_level(self) -> str:
        """
        This function returns the conformance-level (non-numeric part) of this ConformanceLevel
        :return:    the non-numeric part of this ConformanceLevel (A, B, U)
        """
        if self in [ConformanceLevel.PDFA_1A, ConformanceLevel.PDFA_2A]:
            return "A"
        if self in [
            ConformanceLevel.PDFA_1B,
            ConformanceLevel.PDFA_2B,
            ConformanceLevel.PDFA_3B,
        ]:
            return "B"
        if self in [ConformanceLevel.PDFA_2U, ConformanceLevel.PDFA_3U]:
            return "U"
        assert False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of LayoutElement aggregates other LayoutElements
and lays them out underneath each other.
"""
import typing

from borb.io.read.types import Decimal
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.layout_element import LayoutElement


class BlockFlow(LayoutElement):
    """
    This implementation of LayoutElement aggregates other LayoutElements
    and lays them out underneath each other.
    """

    def __init__(self):
        super(BlockFlow, self).__init__()
        self._content: typing.List[LayoutElement] = []

    def add(self, e: LayoutElement) -> "BlockFlow":
        """
        This function adds a LayoutElement to this BlockFlow
        :param e:   the LayoutElement to be added
        :return:    self
        """

        # if the last element of this BlockFlow is an InlineFlow
        # and the new element is also an InlineFlow, just add the two together
        if (
            len(self._content) > 0
            and self._content[-1].__class__.__name__ == "InlineFlow"
            and e.__class__.__name__ == "InlineFlow"
        ):
            self._content[-1].add(e)  # type: ignore[attr-defined]
            return self
        # default behaviour
        self._content.append(e)
        # return
        return self

    def extend(self, es: typing.List[LayoutElement]) -> "BlockFlow":
        """
        This function adds a typing.List of LayoutElement(s) to this BlockFlow
        :param es:   the LayoutElements to be added
        :return:    self
        """
        for e in es:
            self.add(e)
        return self

    def _get_content_box(self, available_space: Rectangle) -> Rectangle:
        height_used: Decimal = Decimal(0)
        height_available: Decimal = available_space.get_height()
        if len(self._content) > 0:
            height_available -= self._content[0].get_margin_top()
        for i, e in enumerate(self._content):
            lbox: Rectangle = e.get_layout_box(
                Rectangle(
                    available_space.get_x(),
                    available_space.get_y(),
                    available_space.get_width(),
                    max(Decimal(0), height_available),
                )
            )
            height_available -= lbox.get_height()
            height_used += lbox.get_height()
            if (i + 1) < len(self._content):
                margin: Decimal = max(
                    e.get_margin_bottom(), self._content[i + 1].get_margin_top()
                )
                height_available -= margin
                height_used += margin
            else:
                height_available -= e.get_margin_bottom()
                height_used += e.get_margin_bottom()

        # return
        return Rectangle(
            available_space.get_x(),
            available_space.get_y() + available_space.get_height() - height_used,
            available_space.get_width(),
            height_used,
        )

    def _paint_content_box(self, page: "Page", content_box: Rectangle) -> None:
        height_available: Decimal = content_box.get_height()
        if len(self._content) > 0:
            height_available -= self._content[0].get_margin_top()
        for i, e in enumerate(self._content):
            e.paint(
                page,
                Rectangle(
                    content_box.get_x(),
                    content_box.get_y(),
                    content_box.get_width(),
                    height_available,
                ),
            )
            lbox: Rectangle = e.get_previous_paint_box()
            height_available -= lbox.get_height()
            if (i + 1) < len(self._content):
                height_available -= max(
                    e.get_margin_bottom(), self._content[i + 1].get_margin_top()
                )
            else:
                height_available -= e.get_margin_bottom()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Modify the current transformation matrix (CTM) by concatenating
    the specified matrix (see 8.3.2, "Coordinate Spaces"). Although the
    operands specify a matrix, they shall be written as six separate
    numbers, not as an array.
"""
import typing
from decimal import Decimal

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.geometry.matrix import Matrix
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class ModifyTransformationMatrix(CanvasOperator):
    """
    Modify the current transformation matrix (CTM) by concatenating
    the specified matrix (see 8.3.2, "Coordinate Spaces"). Although the
    operands specify a matrix, they shall be written as six separate
    numbers, not as an array.
    """

    def __init__(self):
        super().__init__("cm", 6)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the cm operator
        """
        assert isinstance(operands[0], Decimal), "Operand 0 of cm must be a Decimal"
        assert isinstance(operands[1], Decimal), "Operand 1 of cm must be a Decimal"
        assert isinstance(operands[2], Decimal), "Operand 2 of cm must be a Decimal"
        assert isinstance(operands[3], Decimal), "Operand 3 of cm must be a Decimal"
        assert isinstance(operands[4], Decimal), "Operand 4 of cm must be a Decimal"
        assert isinstance(operands[5], Decimal), "Operand 5 of cm must be a Decimal"
        mtx = Matrix.matrix_from_six_values(
            operands[0],
            operands[1],
            operands[2],
            operands[3],
            operands[4],
            operands[5],
        )
        canvas = canvas_stream_processor.get_canvas()
        canvas.graphics_state.ctm = mtx.mul(canvas.graphics_state.ctm)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This implementation of EventListener allows you to search for regular expressions in a PDF Document
"""
import re
import typing
from decimal import Decimal
from functools import cmp_to_key
from typing import List

from borb.pdf.canvas.event.begin_page_event import BeginPageEvent
from borb.pdf.canvas.event.chunk_of_text_render_event import (
    ChunkOfTextRenderEvent,
    LeftToRightComparator,
)
from borb.pdf.canvas.event.end_page_event import EndPageEvent
from borb.pdf.canvas.event.event_listener import Event, EventListener
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.page.page import Page


class PDFMatch:
    """
    This class represents a match of a regular expression in a PDF.
    It has convenience methods to allow the user to extract information about the text that was matched,
    as well as the location (on the page) of the match.
    """

    def __init__(
        self,
        re_match: re.Match,
        glyph_bounding_boxes: typing.List["Rectangle"],
        page_nr: int,
    ):
        super(PDFMatch, self).__init__()
        assert page_nr >= 0
        self._page_nr: int = page_nr
        self._glyph_bounding_boxes: typing.List["Rectangle"] = glyph_bounding_boxes
        self._re_match: re.Match = re_match
        # these fields are kept public to align with the existing python re.Match object
        self.pos = self._re_match.pos
        self.endpos = self._re_match.endpos
        self.lastindex = self._re_match.lastindex
        self.lastgroup = self._re_match.lastgroup
        self.string = self._re_match.string

    def get_bounding_boxes(self) -> typing.List["Rectangle"]:
        """
        This function returns the bounding box(es) that constitute the locations of the glyph(s) that matched the regular expression.
        """
        out: typing.List[Rectangle] = []
        prev_group_of_rectangles: typing.List[Rectangle] = [
            self._glyph_bounding_boxes[0]
        ]
        for i in range(1, len(self._glyph_bounding_boxes)):
            bb: Rectangle = self._glyph_bounding_boxes[i]
            y_delta: Decimal = abs(bb.get_y() - prev_group_of_rectangles[-1].get_y())
            if y_delta > 12:
                max_x = max(
                    [(x.get_x() + x.get_width()) for x in prev_group_of_rectangles]
                )
                min_x = min([x.get_x() for x in prev_group_of_rectangles])
                max_y = max(
                    [(x.get_y() + x.get_height()) for x in prev_group_of_rectangles]
                )
                min_y = min([x.get_y() for x in prev_group_of_rectangles])
                out.append(Rectangle(min_x, min_y, max_x - min_x, max_y - min_y))
                prev_group_of_rectangles.clear()
                prev_group_of_rectangles.append(bb)
                continue
            else:
                prev_group_of_rectangles.append(bb)

        if len(prev_group_of_rectangles) > 0:
            max_x = max([(x.get_x() + x.get_width()) for x in prev_group_of_rectangles])
            min_x = min([x.get_x() for x in prev_group_of_rectangles])
            max_y = max(
                [(x.get_y() + x.get_height()) for x in prev_group_of_rectangles]
            )
            min_y = min([x.get_y() for x in prev_group_of_rectangles])
            out.append(Rectangle(min_x, min_y, max_x - min_x, max_y - min_y))
        return out

    def expand(self, template: typing.AnyStr) -> typing.AnyStr:
        """
        Returns one or more subgroups of the match. If there is a single argument, the result is a single string;
        if there are multiple arguments, the result is a tuple with one item per argument.
        Without arguments, group1 defaults to zero (the whole match is returned).
        If a groupN argument is zero, the corresponding return value is the entire matching string; if it is in the inclusive range [1..99],
        it is the string matching the corresponding parenthesized group.
        If a group number is negative or larger than the number of groups defined in the pattern,
        an IndexError exception is raised. If a group is contained in a part of the pattern that did not match,
        the corresponding result is None. If a group is contained in a part of the pattern that matched multiple times,
        the last match is returned.
        """
        return self._re_match.expand(template)

    def group(self, __group: typing.Union[str, int] = 0) -> typing.AnyStr:
        """
        Return the string obtained by doing backslash substitution on the template string template, as done by the sub() method.
        Escapes such as \n are converted to the appropriate characters, and numeric backreferences (\1, \2)
        and named backreferences (\g<1>, \g<name>) are replaced by the contents of the corresponding group.
        Changed in version 3.5: Unmatched groups are replaced with an empty string.
        """
        return self._re_match.group(__group)

    def __getitem__(self, item):
        """
        This is identical to m.group(g). This allows easier access to an individual group from a match:
        """
        return self._re_match.__getitem__(item)

    def groups(self, default: typing.AnyStr = None) -> typing.Sequence[typing.AnyStr]:
        """
        Return a tuple containing all the subgroups of the match, from 1 up to however many groups are in the pattern.
        The default argument is used for groups that did not participate in the match; it defaults to None.
        """
        return self._re_match.groups(default)  # type: ignore[arg-type]

    def groupdict(
        self, default: typing.AnyStr = None
    ) -> typing.Dict[str, typing.AnyStr]:
        """
        Return a dictionary containing all the named subgroups of the match, keyed by the subgroup name.
        The default argument is used for groups that did not participate in the match; it defaults to None.
        """
        return self._re_match.groupdict(default)  # type: ignore[arg-type]

    def start(self, __group: typing.Union[int, str] = 0) -> int:
        """
        Return the indices of the start of the substring matched by group;
        group defaults to zero (meaning the whole matched substring).
        Return -1 if group exists but did not contribute to the match.
        For a match object m, and a group g that did contribute to the match,
        the substring matched by group g (equivalent to m.group(g)) is m.string[m.start(g):m.end(g)]
        """
        return self._re_match.start(__group)

    def end(self, __group: typing.Union[int, str] = 0) -> int:
        """
        Return the indices of the end of the substring matched by group;
        group defaults to zero (meaning the whole matched substring).
        Return -1 if group exists but did not contribute to the match.
        For a match object m, and a group g that did contribute to the match,
        the substring matched by group g (equivalent to m.group(g)) is m.string[m.start(g):m.end(g)]
        """
        return self._re_match.end(__group)

    def span(self, __group: typing.Union[int, str] = 0) -> typing.Tuple[int, int]:
        """
        For a match m, return the 2-tuple (m.start(group), m.end(group)).
        Note that if group did not contribute to the match, this is (-1, -1). group defaults to zero, the entire match.
        """
        return self._re_match.span(__group)


class RegularExpressionTextExtraction(EventListener):
    """
    This implementation of EventListener allows you to search for regular expressions in a PDF Document
    """

    def __init__(self, regular_expression):
        self._regular_expression = regular_expression
        self._text_render_info_events_per_page: typing.Dict[
            int, typing.List[ChunkOfTextRenderEvent]
        ] = {}
        self._matches_per_page: typing.Dict[int, typing.List[PDFMatch]] = {}
        self._text_per_page: typing.Dict[int, str] = {}
        self._current_page: int = -1

    def _event_occurred(self, event: Event) -> None:
        if isinstance(event, ChunkOfTextRenderEvent):
            self._render_text(event)
        if isinstance(event, BeginPageEvent):
            self._begin_page(event.get_page())
        if isinstance(event, EndPageEvent):
            self._end_page(event.get_page())

    def _render_text(self, text_render_info: ChunkOfTextRenderEvent):

        # init if needed
        if self._current_page not in self._text_render_info_events_per_page:
            self._text_render_info_events_per_page[self._current_page] = []

        # append TextRenderInfo
        for e in text_render_info.split_on_glyphs():
            self._text_render_info_events_per_page[self._current_page].append(e)

    def _begin_page(self, page: Page):
        self._current_page += 1

    def _end_page(self, page: Page):

        # get ChunkOfTextRenderEvent objects on page
        tris: typing.List[ChunkOfTextRenderEvent] = (
            self._text_render_info_events_per_page[self._current_page]
            if self._current_page in self._text_render_info_events_per_page
            else []
        )

        # remove no-op
        tris = [x for x in tris if len(x.get_text().replace(" ", "")) != 0]

        # skip empty
        if len(tris) == 0:
            return

        # sort according to comparator
        sorted(tris, key=cmp_to_key(LeftToRightComparator.cmp))

        poss = []

        # iterate over the ChunkOfTextRenderEvent objects to get the text
        last_baseline_bottom = tris[0].get_baseline().y
        last_baseline_right = tris[0].get_baseline().x
        text = ""
        for t in tris:

            chunk_of_text_bounding_box: typing.Optional[
                Rectangle
            ] = t.get_previous_layout_box()
            assert chunk_of_text_bounding_box is not None

            # add newline if needed
            if abs(t.get_baseline().y - last_baseline_bottom) > 10 and len(text) > 0:
                if text.endswith(" "):
                    text = text[0:-1]
                text += "\n"
                text += t.get_text()
                last_baseline_right = (
                    chunk_of_text_bounding_box.get_x()
                    + chunk_of_text_bounding_box.get_width()
                )
                last_baseline_bottom = t.get_baseline().y
                poss.append(len(text))
                continue

            # check text
            if t.get_text().startswith(" ") or text.endswith(" "):
                text += t.get_text()
                last_baseline_right = (
                    chunk_of_text_bounding_box.get_x()
                    + chunk_of_text_bounding_box.get_width()
                )
                poss.append(len(text))
                continue

            # add space if needed
            delta = abs(last_baseline_right - chunk_of_text_bounding_box.get_x())
            space_width = round(t.get_space_character_width_estimate_in_user_space(), 1)
            text += " " if (space_width * Decimal(0.90) < delta) else ""

            # normal append
            text += t.get_text()
            last_baseline_right = (
                chunk_of_text_bounding_box.get_x()
                + chunk_of_text_bounding_box.get_width()
            )
            poss.append(len(text))
            continue

        # attempt to match
        for m in re.finditer(self._regular_expression, text):
            tri_start_index = len(
                [x for x in poss if x <= m.start()]
            )  # here we use '<=' because poss contains the number of characters we have seen at position x, with x included
            tri_stop_index = len([x for x in poss if x < m.end()])

            # set up list if we don't have information yet for this Page
            if self._current_page not in self._matches_per_page:
                self._matches_per_page[self._current_page] = []

            # extend collection
            self._matches_per_page[self._current_page].append(
                PDFMatch(
                    m,
                    [
                        x.get_previous_layout_box()  # type: ignore [misc]
                        for x in tris[tri_start_index : (tri_stop_index + 1)]
                    ],
                    self._current_page,
                )
            )

    def get_matches_for_page(self, page_number: int) -> List[PDFMatch]:
        """
        This function returns a typing.List[PDFMatch] matching the regular expression, on a given page
        """
        if page_number not in self._matches_per_page:
            return []
        return self._matches_per_page[page_number]

    def get_text_for_page(self, page_nr: int) -> str:
        """
        This function returns all text on a given page
        """
        return self._text_per_page[page_nr] if page_nr in self._text_per_page else ""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Show one or more text strings, allowing individual glyph positioning. Each
element of array shall be either a string or a number.
"""
import typing
from decimal import Decimal
from typing import List

from borb.io.read.types import AnyPDFType, Name, String
from borb.pdf.canvas.event.chunk_of_text_render_event import ChunkOfTextRenderEvent
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class ShowTextWithGlyphPositioning(CanvasOperator):
    """
    Show one or more text strings, allowing individual glyph positioning. Each
    element of array shall be either a string or a number. If the element is a
    string, this operator shall show the string. If it is a number, the operator
    shall adjust the text position by that amount; that is, it shall translate the
    text matrix, T . The number shall be expressed in thousandths of a unit
    mof text space (see 9.4.4, "Text Space Details"). This amount shall be
    subtracted from the current horizontal or vertical coordinate, depending
    on the writing mode. In the default coordinate system, a positive
    adjustment has the effect of moving the next glyph painted either to the
    left or down by the given amount. Figure 46 shows an example of the
    effect of passing offsets to TJ.
    """

    def __init__(self):
        super().__init__("TJ", 1)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the TJ operator
        """

        assert isinstance(operands[0], List), "Operand 0 of TJ must be a List"
        canvas = canvas_stream_processor.get_canvas()

        # handle Font being a Name (optimization)
        assert canvas.graphics_state.font is not None
        font_name: typing.Optional[Name] = None
        if isinstance(canvas.graphics_state.font, Name):
            font_name = canvas.graphics_state.font
            canvas.graphics_state.font = canvas_stream_processor.get_resource(
                "Font", canvas.graphics_state.font
            )

        for i in range(0, len(operands[0])):
            obj = operands[0][i]

            # display string
            if isinstance(obj, String):
                assert isinstance(obj, String)
                tri = ChunkOfTextRenderEvent(canvas.graphics_state, obj)
                # render
                for l in event_listeners:
                    l._event_occurred(tri)
                # update text rendering location
                canvas.graphics_state.text_matrix[2][0] += tri.get_baseline().width
                continue

            # adjust
            if isinstance(obj, Decimal):
                assert isinstance(obj, Decimal)
                gs = canvas.graphics_state
                adjust_unscaled = obj
                adjust_scaled = (
                    -adjust_unscaled
                    * Decimal(0.001)
                    * gs.font_size
                    * (gs.horizontal_scaling / 100)
                )
                gs.text_matrix[2][0] -= adjust_scaled

        # restore
        if font_name is not None:
            canvas.graphics_state.font = font_name



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Begin a new subpath by moving the current point to
coordinates (x, y), omitting any connecting line segment. If
the previous path construction operator in the current path
was also m, the new m overrides it; no vestige of the
previous m operation remains in the path.
"""
import typing
from decimal import Decimal

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.geometry.line_segment import LineSegment
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class BeginSubpath(CanvasOperator):
    """
    Begin a new subpath by moving the current point to
    coordinates (x, y), omitting any connecting line segment. If
    the previous path construction operator in the current path
    was also m, the new m overrides it; no vestige of the
    previous m operation remains in the path.
    """

    def __init__(self):
        super().__init__("m", 2)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the m operator
        """
        assert isinstance(
            operands[0], Decimal
        ), "operand 0 of m operator must be of type Decimal"
        assert isinstance(
            operands[1], Decimal
        ), "operand 1 of m operator must be of type Decimal"

        # get graphic state
        canvas = canvas_stream_processor.get_canvas()
        gs = canvas.graphics_state

        # start empty subpath
        gs.path.append(LineSegment(operands[0], operands[1], operands[0], operands[1]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    In computer science, a disjoint-set data structure, also called a union–find data structure or merge–find set,
    is a data structure that stores a collection of disjoint (non-overlapping) sets.
"""
import typing
from typing import Any, List


class disjointset:
    """
    In computer science, a disjoint-set data structure, also called a union–find data structure or merge–find set,
    is a data structure that stores a collection of disjoint (non-overlapping) sets.
    Equivalently, it stores a partition of a set into disjoint subsets.
    It provides operations for adding new sets, merging sets (replacing them by their union),
    and finding a representative member of a set.
    The last operation allows to find out efficiently if any two elements are in the same or different sets.
    """

    def __init__(self):
        self._parents = {}
        self._ranks = {}

    def find(self, x: Any) -> Any:
        """
        Find the root of an element in this disjointset
        """
        if self._parents[x] == x:
            return x
        else:
            return self.find(self._parents[x])

    def union(self, x: Any, y: Any) -> "disjointset":
        """
        Mark two elements in this disjointset as equivalent,
        propagating the equivalence throughout the disjointset
        """
        x_parent = self.find(x)
        y_parent = self.find(y)
        if x_parent is y_parent:
            return self
        if self._ranks[x_parent] > self._ranks[y_parent]:
            self._parents[y_parent] = x_parent
        elif self._ranks[y_parent] > self._ranks[x_parent]:
            self._parents[x_parent] = y_parent
        else:
            self._parents[y_parent] = x_parent
            self._ranks[x_parent] += 1
        return self

    def add(self, x: Any) -> "disjointset":
        """
        Add an element to this disjointset
        """
        self._parents[x] = x
        self._ranks[x] = 0
        return self

    def pop(self, x: Any) -> "disjointset":
        """
        Remove an element from this disjointset
        """
        raise NotImplementedError()

    def sets(self) -> List[List[Any]]:
        """
        This function returns all equivalence sets in this disjointset
        """
        cluster_parents: typing.Dict[Any, Any] = {}
        for x, _ in self._parents.items():
            p = self.find(x)
            if p not in cluster_parents:
                cluster_parents[p] = []
            cluster_parents[p].append(x)
        return [v for k, v in cluster_parents.items()]

    def __len__(self):
        return len(self._parents)

    def __contains__(self, item):
        return item in self._parents

    def __iter__(self):
        return self._parents.__iter__()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Set the horizontal scaling, Th , to (scale ÷ 100). scale shall be a number
specifying the percentage of the normal width. Initial value: 100 (normal
width).
"""
import typing
from decimal import Decimal

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class SetHorizontalScaling(CanvasOperator):
    """
    Set the horizontal scaling, Th , to (scale ÷ 100). scale shall be a number
    specifying the percentage of the normal width. Initial value: 100 (normal
    width).
    """

    def __init__(self):
        super().__init__("Tz", 1)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the Tz operator
        """
        assert isinstance(operands[0], Decimal), "Operand 0 of Tz must be a Decimal"
        canvas = canvas_stream_processor.get_canvas()
        canvas.graphics_state.horizontal_scaling = operands[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of ReadBaseTransformer is responsible for reading Reference objects
e.g. 97 0 R
"""
import io
import logging
import typing
from typing import Any, Optional, Union

from borb.io.read.transformer import ReadTransformerState, Transformer
from borb.io.read.types import AnyPDFType, Reference
from borb.pdf.canvas.event.event_listener import EventListener
from borb.pdf.xref.xref import XREF

logger = logging.getLogger(__name__)


class ReferenceTransformer(Transformer):
    """
    This implementation of ReadBaseTransformer is responsible for reading Reference objects
    e.g. 97 0 R
    """

    def __init__(self):
        super(ReferenceTransformer, self).__init__()
        self._cache: typing.Dict[Reference, AnyPDFType] = {}
        self._cache_hits: int = 0
        self._cache_fails: int = 0

    def can_be_transformed(
        self, object: Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO, AnyPDFType]
    ) -> bool:
        """
        This function returns True if the object to be converted represents a Reference
        """
        return isinstance(object, Reference)

    def transform(
        self,
        object_to_transform: Union[io.BufferedIOBase, io.RawIOBase, AnyPDFType],
        parent_object: Any,
        context: Optional[ReadTransformerState] = None,
        event_listeners: typing.List[EventListener] = [],
    ) -> Any:
        """
        This function reads a Reference from a byte stream
        """
        # fmt: off
        assert isinstance(object_to_transform, Reference), "object_to_transform must be of type Reference"
        # fmt: on

        # check for circular reference
        assert context is not None
        if object_to_transform in context.indirect_reference_chain:
            return None

        # lookup in cache
        ref_from_cache = self._cache.get(object_to_transform, None)
        if ref_from_cache is not None:
            self._cache_hits += 1
            # check linkage
            if ref_from_cache.get_parent() is None:  # type: ignore[union-attr]
                ref_from_cache.set_parent(parent_object)  # type: ignore[union-attr]
                return ref_from_cache
            # copy because of linkage
            if ref_from_cache.get_parent() != parent_object:  # type: ignore[union-attr]
                ref_from_cache_copy = ref_from_cache  # TODO
                ref_from_cache_copy.set_parent(parent_object)  # type: ignore[union-attr]
                return ref_from_cache_copy

        self._cache_fails += 1
        logger.debug(
            "ref. cache hits: %d, fails: %d, ratio %f"
            % (
                self._cache_hits,
                self._cache_fails,
                self._cache_hits / (self._cache_hits + self._cache_fails),
            )
        )

        # lookup xref
        # fmt: off
        assert (context.root_object is not None), "context.root_object must be defined to read Reference objects"
        assert context.root_object["XRef"] is not None, "XREF must be defined to read Reference objects"
        assert isinstance(context.root_object["XRef"], XREF), "XREF must be defined to read Reference objects"
        assert (context.tokenizer is not None), "context.tokenizer must be defined to read Reference objects"
        assert (context.source is not None), "context.source must be defined to read Reference objects"
        # fmt: on

        xref = context.root_object["XRef"]
        src = context.source
        tok = context.tokenizer

        # get reference
        referenced_object = xref.get_object(object_to_transform, src, tok)
        if referenced_object is None:
            return None

        # transform
        assert referenced_object is not None
        context.indirect_reference_chain.add(object_to_transform)
        transformed_referenced_object = self.get_root_transformer().transform(
            referenced_object, parent_object, context, event_listeners
        )
        context.indirect_reference_chain.remove(object_to_transform)

        # update cache
        if transformed_referenced_object is not None:
            self._cache[object_to_transform] = transformed_referenced_object

        # set reference
        try:
            transformed_referenced_object.set_reference(object_to_transform)
        except:
            logger.debug(
                "Unable to set reference on object %s"
                % str(transformed_referenced_object)
            )
            pass

        # return
        return transformed_referenced_object



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
(PDF 1.1) Begin a compatibility section. Unrecognized operators (along with
their operands) shall be ignored without error until the balancing EX operator
is encountered.
"""

import typing

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class EndCompatibilitySection(CanvasOperator):
    """
    (PDF 1.1) Begin a compatibility section. Unrecognized operators (along with
    their operands) shall be ignored without error until the balancing EX operator
    is encountered.
    """

    def __init__(self):
        super().__init__("EX", 0)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the EX operator
        """
        canvas_stream_processor.get_canvas().in_compatibility_section = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Close the current subpath by appending a straight line
segment from the current point to the starting point of the
subpath. If the current subpath is already closed, h shall do
nothing.
"""

import typing

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.geometry.line_segment import LineSegment
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class CloseSubpath(CanvasOperator):
    """
    Close the current subpath by appending a straight line
    segment from the current point to the starting point of the
    subpath. If the current subpath is already closed, h shall do
    nothing.

    This operator terminates the current subpath. Appending
    another segment to the current path shall begin a new
    subpath, even if the new segment begins at the endpoint
    reached by the h operation.
    """

    def __init__(self):
        super().__init__("h", 0)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the h operator
        """

        # get graphic state
        canvas = canvas_stream_processor.get_canvas()
        gs = canvas.graphics_state

        # path is empty
        if len(gs.path) == 0:
            return

        # first point in subpath
        x0 = gs.path[0].x0
        y0 = gs.path[0].y0

        # last point in subpath
        xn = gs.path[-1].x1
        yn = gs.path[-1].y1

        # append straight line segment
        gs.path.append(LineSegment(x0, y0, xn, yn))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
End a marked-content sequence begun by a BMC or BDC operator.
"""

import typing

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class EndMarkedContent(CanvasOperator):
    """
    End a marked-content sequence begun by a BMC or BDC operator.
    """

    def __init__(self):
        super().__init__("EMC", 0)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the EMC operator
        """
        canvas = canvas_stream_processor.get_canvas()
        assert len(canvas.marked_content_stack) > 0
        canvas.marked_content_stack.pop(-1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A watermark annotation (PDF 1.6) shall be used to represent graphics that shall be printed at a fixed size and
position on a page, regardless of the dimensions of the printed page. The FixedPrint entry of a watermark
annotation dictionary (see Table 190) shall be a dictionary that contains values for specifying the size and
position of the annotation (see Table 191).
"""
from borb.pdf.canvas.layout.annotation.annotation import Annotation


class WatermarkAnnotation(Annotation):
    """
    A watermark annotation (PDF 1.6) shall be used to represent graphics that shall be printed at a fixed size and
    position on a page, regardless of the dimensions of the printed page. The FixedPrint entry of a watermark
    annotation dictionary (see Table 190) shall be a dictionary that contains values for specifying the size and
    position of the annotation (see Table 191).
    """

    def __init__(self):
        super(WatermarkAnnotation, self).__init__()
        # TODO



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
3D annotations (PDF 1.6) are the means by which 3D artwork shall be represented in a PDF document.
Table 298 shows the entries specific to a 3D annotation dictionary. Table 164 describes the entries common to
all annotation dictionaries.

In addition to these entries, a 3D annotation shall provide an appearance stream in its AP entry (see Table 164)
that has a normal appearance (the N entry in Table 168). This appearance may be used by applications that do
not support 3D annotations and by all applications for the initial display of the annotation.
"""
from borb.pdf.canvas.layout.annotation.annotation import Annotation


class ThreeDAnnotation(Annotation):
    """
    3D annotations (PDF 1.6) are the means by which 3D artwork shall be represented in a PDF document.
    Table 298 shows the entries specific to a 3D annotation dictionary. Table 164 describes the entries common to
    all annotation dictionaries.

    In addition to these entries, a 3D annotation shall provide an appearance stream in its AP entry (see Table 164)
    that has a normal appearance (the N entry in Table 168). This appearance may be used by applications that do
    not support 3D annotations and by all applications for the initial display of the annotation.
    """

    def __init__(self):
        super(ThreeDAnnotation, self).__init__()
        # TODO



# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""
The language that shall be used in a type 4 function contains expressions involving integers, real numbers, and
boolean values only. There shall be no composite data structures such as strings or arrays, no procedures, and
no variables or names. Table 42 lists the operators that can be used in this type of function. (For more
information on these operators, see Appendix B of the PostScript Language Reference, Third Edition.)
Although the semantics are those of the corresponding PostScript operators, a full PostScript interpreter is not
required.
"""
import typing
from decimal import Decimal
from math import atan, ceil, cos, degrees, exp, floor, log, sin, sqrt


class PostScriptEval:
    """
    The language that shall be used in a type 4 function contains expressions involving integers, real numbers, and
    boolean values only. There shall be no composite data structures such as strings or arrays, no procedures, and
    no variables or names. Table 42 lists the operators that can be used in this type of function. (For more
    information on these operators, see Appendix B of the PostScript Language Reference, Third Edition.)
    Although the semantics are those of the corresponding PostScript operators, a full PostScript interpreter is not
    required.
    """

    @staticmethod
    def evaluate(s: str, args: typing.List[Decimal]) -> typing.List[Decimal]:
        """
        This function evaluates a postscript str, using args as the (initial) stack.
        This function returns a typing.List[Decimal], or throws an assertion error
        """
        stk: typing.List[typing.Union[Decimal, bool]] = [x for x in args]
        ops: typing.List[str] = [
            "abs",
            "add",
            "and",
            "atan",
            "bitshift",  # TODO
            "ceiling",
            "copy",
            "cos",
            "cvi",
            "cvr",
            "div",
            "dup",
            "eq",
            "exch",
            "exp",
            "false",
            "floor",
            "ge",
            "gt",
            "idiv",
            "index",  # TODO
            "le",
            "ln",
            "log",
            "lt",
            "mod",
            "mul",
            "ne",
            "neg",
            "not",
            "or",
            "pop",
            "roll",  # TODO
            "round",
            "sin",
            "sqrt",
            "sub",
            "true",
            "truncate",  # TODO
            "xor",
        ]

        i: int = 0
        while i < len(s):

            # whitespace
            if s[i] in " \n\t":
                i += 1
                continue

            # brackets
            if s[i] == "{" or s[i] == "}":
                i += 1
                continue

            # operand
            if s[i] in "0123456789.-":
                operand: str = ""
                while i < len(s) and s[i] in "0123456789.-":
                    operand += s[i]
                    i += 1
                stk.append(Decimal(operand))
                continue

            # operator
            if any([x.startswith(s[i]) for x in ops]):
                operator: str = ""
                while i < len(s) and s[i] in "abcdefghijklmnopqrstuvwxyz":
                    operator += s[i]
                    i += 1
                if operator not in ops:
                    assert False, "Unknown operator %s in postscript str" % operator

                x: typing.Optional[typing.Union[Decimal, bool]] = None
                y: typing.Optional[typing.Union[Decimal, bool]] = None

                # abs
                if operator == "abs":
                    # fmt: off
                    assert (len(stk) >= 1), "Unable to apply operator abs, stack underflow"
                    assert isinstance(stk[-1], Decimal), "Unable to apply operator abs, arg 1 must be of type Decimal"
                    # fmt: on
                    x = stk[-1]
                    stk.pop(len(stk) - 1)
                    stk.append(abs(x))
                    continue
                # add
                if operator == "add":
                    # fmt: off
                    assert (len(stk) >= 2), "Unable to apply operator add, stack underflow"
                    assert isinstance(stk[-1], Decimal), "Unable to apply operator add, arg 1 must be of type Decimal"
                    assert isinstance(stk[-2], Decimal), "Unable to apply operator add, arg 2 must be of type Decimal"
                    # fmt: on
                    x = stk[-1]
                    y = stk[-2]
                    stk.pop(len(stk) - 1)
                    stk.pop(len(stk) - 1)
                    stk.append(x + y)
                    continue
                # and
                if operator == "and":
                    # fmt: off
                    assert (len(stk) >= 2), "Unable to apply operator and, stack underflow"
                    assert isinstance(stk[-1], bool), "Unable to apply operator and, arg 1 must be of type bool"
                    assert isinstance(stk[-2], bool), "Unable to apply operator and, arg 2 must be of type bool"
                    # fmt: on
                    x = stk[-1]
                    y = stk[-2]
                    stk.pop(len(stk) - 1)
                    stk.pop(len(stk) - 1)
                    stk.append(x and y)
                    continue
                # atan
                if operator == "atan":
                    # fmt: off
                    assert (len(stk) >= 1), "Unable to apply operator atan, stack underflow"
                    assert isinstance(stk[-1], Decimal), "Unable to apply operator atan, arg 1 must be of type Decimal"
                    # fmt: on
                    x = stk[-1]
                    stk.pop(len(stk) - 1)
                    stk.append(Decimal(atan(x)))
                    continue
                # ceiling
                if operator == "ceiling":
                    # fmt: off
                    assert (len(stk) >= 1), "Unable to apply operator ceiling, stack underflow"
                    assert isinstance(stk[-1], Decimal), "Unable to apply operator ceiling, arg 1 must be of type Decimal"
                    # fmt: on
                    x = stk[-1]
                    stk.pop(len(stk) - 1)
                    stk.append(Decimal(ceil(x)))
                    continue
                # cos
                if operator == "cos":
                    # fmt: off
                    assert (len(stk) >= 1), "Unable to apply operator cos, stack underflow"
                    assert isinstance(stk[-1], Decimal), "Unable to apply operator cos, arg 1 must be of type Decimal"
                    # fmt: on
                    x = stk[-1]
                    stk.pop(len(stk) - 1)
                    stk.append(Decimal(cos(degrees(x))))
                    continue
                # cvi
                if operator == "cvi":
                    # fmt: off
                    assert (len(stk) >= 1), "Unable to apply operator cvi, stack underflow"
                    assert isinstance(stk[-1], Decimal), "Unable to apply operator cvi, arg 1 must be of type Decimal"
                    # fmt: on
                    x = stk[-1]
                    stk.pop(len(stk) - 1)
                    stk.append(Decimal(int(x)))
                    continue
                # cvr
                if operator == "cvr":
                    # fmt: off
                    assert (len(stk) >= 1), "Unable to apply operator cvr, stack underflow"
                    # fmt: on
                # div
                if operator == "div":
                    # fmt: off
                    assert (len(stk) >= 2), "Unable to apply operator div, stack underflow"
                    assert isinstance(stk[-1], Decimal), "Unable to apply operator div, arg 1 must be of type Decimal"
                    assert isinstance(stk[-2], Decimal), "Unable to apply operator div, arg 2 must be of type Decimal"
                    assert stk[-1] != Decimal(0), "Unable to apply operator div, arg1 must not be 0"
                    # fmt: on
                    x = stk[-1]
                    y = stk[-2]
                    stk.pop(len(stk) - 1)
                    stk.pop(len(stk) - 1)
                    stk.append(y / x)
                    continue
                # dup
                if operator == "dup":
                    # fmt: off
                    assert (len(stk) >= 1), "Unable to apply operator dup, stack underflow"
                    # fmt: on
                    stk.append(stk[-1])
                    continue
                # eq
                if operator == "eq":
                    assert len(stk) >= 2, "Unable to apply operator eq, stack underflow"
                    x = stk[-1]
                    y = stk[-2]
                    stk.pop(len(stk) - 1)
                    stk.pop(len(stk) - 1)
                    stk.append(x == y)
                    continue
                # exch
                if operator == "exch":
                    # fmt: off
                    assert (len(stk) >= 2), "Unable to apply operator exch, stack underflow"
                    # fmt: on
                    x = stk[-1]
                    y = stk[-2]
                    stk.pop(len(stk) - 1)
                    stk.pop(len(stk) - 1)
                    stk.append(x)
                    stk.append(y)
                    continue
                # exp
                if operator == "exp":
                    # fmt: off
                    assert len(stk) >= 1, "Unable to apply operator exp, stack underflow"
                    # fmt: on
                    x = stk[-1]
                    assert isinstance(x, Decimal)
                    stk.pop(len(stk) - 1)
                    stk.append(Decimal(exp(x)))
                    continue
                # false
                if operator == "false":
                    stk.append(False)
                    continue
                # floor
                if operator == "floor":
                    # fmt: off
                    assert len(stk) >= 1, "Unable to apply operator floor, stack underflow"
                    # fmt: on
                    x = stk[-1]
                    assert isinstance(x, Decimal)
                    stk.pop(len(stk) - 1)
                    stk.append(Decimal(floor(x)))
                    continue
                # ge
                if operator == "ge":
                    assert len(stk) >= 2, "Unable to apply operator ge, stack underflow"
                    x = stk[-1]
                    y = stk[-2]
                    assert isinstance(x, Decimal)
                    assert isinstance(y, Decimal)
                    stk.pop(len(stk) - 1)
                    stk.pop(len(stk) - 1)
                    stk.append(y >= x)
                    continue
                # gt
                if operator == "gt":
                    assert len(stk) >= 2, "Unable to apply operator gt, stack underflow"
                    x = stk[-1]
                    y = stk[-2]
                    assert isinstance(x, Decimal)
                    assert isinstance(y, Decimal)
                    stk.pop(len(stk) - 1)
                    stk.pop(len(stk) - 1)
                    stk.append(y > x)
                    continue
                # idiv
                if operator == "idiv":
                    # fmt: off
                    assert len(stk) >= 2, "Unable to apply operator idiv, stack underflow"
                    # fmt: on
                    x = stk[-1]
                    y = stk[-2]
                    assert isinstance(x, Decimal)
                    assert isinstance(y, Decimal)
                    stk.pop(len(stk) - 1)
                    stk.pop(len(stk) - 1)
                    # fmt: off
                    assert x != Decimal(0), "Unable to apply operator idiv, division by zero"
                    # fmt: on
                    stk.append(Decimal(int(y / x)))
                    continue
                # le
                if operator == "le":
                    assert len(stk) >= 2, "Unable to apply operator le, stack underflow"
                    x = stk[-1]
                    y = stk[-2]
                    assert isinstance(x, Decimal)
                    assert isinstance(y, Decimal)
                    stk.pop(len(stk) - 1)
                    stk.pop(len(stk) - 1)
                    stk.append(y <= x)
                    continue
                # ln
                if operator == "ln":
                    assert len(stk) >= 1, "Unable to apply operator ln, stack underflow"
                    x = stk[-1]
                    assert isinstance(x, Decimal)
                    stk.pop(len(stk) - 1)
                    stk.append(Decimal(log(x)))
                    continue
                # log
                if operator == "log":
                    # fmt: off
                    assert len(stk) >= 1, "Unable to apply operator log, stack underflow"
                    # fmt: on
                    x = stk[-1]
                    assert isinstance(x, Decimal)
                    stk.pop(len(stk) - 1)
                    stk.append(Decimal(log(x, Decimal(10))))
                    continue
                # lt
                if operator == "lt":
                    assert len(stk) >= 2, "Unable to apply operator lt, stack underflow"
                    x = stk[-1]
                    y = stk[-2]
                    assert isinstance(x, Decimal)
                    assert isinstance(y, Decimal)
                    stk.pop(len(stk) - 1)
                    stk.pop(len(stk) - 1)
                    stk.append(y < x)
                    continue
                # mod
                if operator == "mod":
                    # fmt: off
                    assert len(stk) >= 2, "Unable to apply operator mod, stack underflow"
                    # fmt: on
                    x = stk[-1]
                    y = stk[-2]
                    assert isinstance(x, Decimal)
                    assert isinstance(y, Decimal)
                    stk.pop(len(stk) - 1)
                    stk.pop(len(stk) - 1)
                    # fmt: off
                    assert y != Decimal(0), "Unable to apply operator mod, division by zero"
                    # fmt: on
                    stk.append(Decimal(int(y) % int(x)))
                    continue
                # mul
                if operator == "mul":
                    # fmt: off
                    assert len(stk) >= 2, "Unable to apply operator mul, stack underflow"
                    # fmt: on
                    x = stk[-1]
                    y = stk[-2]
                    assert isinstance(x, Decimal)
                    assert isinstance(y, Decimal)
                    stk.pop(len(stk) - 1)
                    stk.pop(len(stk) - 1)
                    stk.append(y * x)
                    continue
                # ne
                if operator == "ne":
                    assert len(stk) >= 2, "Unable to apply operator ne, stack underflow"
                    x = stk[-1]
                    y = stk[-2]
                    stk.pop(len(stk) - 1)
                    stk.pop(len(stk) - 1)
                    stk.append(y != x)
                    continue
                # neg
                if operator == "neg":
                    # fmt: off
                    assert len(stk) >= 1, "Unable to apply operator neg, stack underflow"
                    # fmt: on
                    x = stk[-1]
                    assert isinstance(x, Decimal)
                    stk.pop(len(stk) - 1)
                    stk.append(-x)
                    continue
                # not
                if operator == "not":
                    # fmt: off
                    assert len(stk) >= 1, "Unable to apply operator not, stack underflow"
                    # fmt: on
                    x = stk[-1]
                    assert isinstance(x, bool)
                    stk.pop(len(stk) - 1)
                    stk.append(not x)
                    continue
                # or
                if operator == "or":
                    assert len(stk) >= 2, "Unable to apply operator or, stack underflow"
                    x = stk[-1]
                    y = stk[-2]
                    assert isinstance(x, bool)
                    assert isinstance(y, bool)
                    stk.pop(len(stk) - 1)
                    stk.pop(len(stk) - 1)
                    stk.append(y or x)
                    continue
                # pop
                if operator == "pop":
                    # fmt: off
                    assert len(stk) >= 1, "Unable to apply operator pop, stack underflow"
                    # fmt: on
                    stk.pop(-1)
                    continue
                # round
                if operator == "round":
                    # fmt: off
                    assert len(stk) >= 1, "Unable to apply operator round, stack underflow"
                    # fmt: on
                    x = stk[-1]
                    assert isinstance(x, Decimal)
                    stk.pop(len(stk) - 1)
                    stk.append(Decimal(round(x)))
                    continue
                # sin
                if operator == "sin":
                    # fmt: off
                    assert len(stk) >= 1, "Unable to apply operator sin, stack underflow"
                    # fmt: on
                    x = stk[-1]
                    assert isinstance(x, Decimal)
                    stk.pop(len(stk) - 1)
                    stk.append(Decimal(sin(degrees(x))))
                    continue
                # sqrt
                if operator == "sqrt":
                    # fmt: off
                    assert len(stk) >= 1, "Unable to apply operator sqrt, stack underflow"
                    # fmt: on
                    x = stk[-1]
                    assert isinstance(x, Decimal)
                    stk.pop(len(stk) - 1)
                    stk.append(Decimal(sqrt(x)))
                    continue
                # sub
                if operator == "sub":
                    # fmt: off
                    assert len(stk) >= 2, "Unable to apply operator sub, stack underflow"
                    # fmt: on
                    x = stk[-1]
                    y = stk[-2]
                    assert isinstance(x, Decimal)
                    assert isinstance(y, Decimal)
                    stk.pop(len(stk) - 1)
                    stk.pop(len(stk) - 1)
                    stk.append(y - x)
                    continue
                # true
                if operator == "true":
                    stk.append(True)
                    continue
                # xor
                if operator == "xor":
                    # fmt: off
                    assert len(stk) >= 2, "Unable to apply operator xor, stack underflow"
                    # fmt: on
                    x = stk[-1]
                    y = stk[-2]
                    assert isinstance(x, bool)
                    assert isinstance(y, bool)
                    stk.pop(len(stk) - 1)
                    stk.pop(len(stk) - 1)
                    stk.append(x or y and not (x and y))
                    continue

            # unknown, advance by 1
            i += 1

        # check type(s)
        out: typing.List[Decimal] = []
        for x in stk:
            assert isinstance(x, Decimal)
            out.append(x)

        # return
        return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
(PDF 1.2) Decompresses data encoded using the zlib/deflate
compression method, reproducing the original text or binary
data.
"""
import copy
import zlib
from typing import List


class FlateDecode:
    """
    (PDF 1.2) Decompresses data encoded using the zlib/deflate
    compression method, reproducing the original text or binary
    data.
    """

    @staticmethod
    def decode(
        bytes_in: bytes,
        predictor: int = 1,
        bits_per_component: int = 8,
        columns: int = 1,
    ) -> bytes:
        """
        Decompresses data encoded using the zlib/deflate
        compression method
        """

        # trivial case
        if len(bytes_in) == 0:
            return bytes_in

        # check /Predictor
        # fmt: off
        assert predictor in [1, 2, 10, 11, 12, 13, 14, 15,], "Illegal argument exception. predictor must be in [1, 2, 10, 11, 12, 13, 14, 15]."
        # fmt: on

        # check /BitsPerComponent
        # fmt: off
        assert bits_per_component in [1, 2, 4, 8], "Illegal argument exception. bits_per_component must be in [1, 2, 4, 8]."
        # fmt: on

        # initial transform
        bytes_after_zlib = zlib.decompress(bytes_in, bufsize=4092)

        # LZW and Flate encoding compress more compactly if their input data is highly predictable. One way of
        # increasing the predictability of many continuous-tone sampled images is to replace each sample with the
        # difference between that sample and a predictor function applied to earlier neighboring samples. If the predictor
        # function works well, the postprediction data clusters toward 0.
        # PDF supports two groups of Predictor functions. The first, the TIFF group, consists of the single function that is
        # Predictor 2 in the TIFF 6.0 specification.

        # check predictor
        if predictor == 1:
            return bytes_after_zlib

        # set up everything to do PNG prediction
        bytes_per_row: int = int((columns * bits_per_component + 7) / 8)
        bytes_per_pixel = int(bits_per_component / 8)

        current_row: List[int] = [0 for _ in range(0, bytes_per_row)]
        prior_row: List[int] = [0 for _ in range(0, bytes_per_row)]
        number_of_rows = int(len(bytes_after_zlib) / bytes_per_row)

        # easy case
        bytes_after_predictor = [int(x) for x in bytes_after_zlib]
        if predictor == 2:
            if bits_per_component == 8:
                for row in range(0, number_of_rows):
                    row_start_index = row * bytes_per_row
                    for col in range(1, bytes_per_row):
                        bytes_after_predictor[row_start_index + col] = (
                            bytes_after_predictor[row_start_index + col]
                            + bytes_after_predictor[row_start_index + col - 1]
                        ) % 256
                return bytes([(int(x) % 256) for x in bytes_after_predictor])

        # harder cases
        bytes_after_predictor = []
        pos = 0
        while pos + bytes_per_row <= len(bytes_after_zlib):

            # Read the filter type byte and a row of data
            filter_type = bytes_after_zlib[pos]
            pos += 1

            current_row = [x for x in bytes_after_zlib[pos : pos + bytes_per_row]]
            pos += bytes_per_row

            # PNG_FILTER_NONE
            if filter_type == 0:
                # DO NOTHING
                pass

            # PNG_FILTER_SUB
            # Predicts the same as the sample to the left
            if filter_type == 1:
                for i in range(bytes_per_pixel, bytes_per_row):
                    current_row[i] = (
                        current_row[i] + current_row[i - bytes_per_pixel]
                    ) % 256

            # PNG_FILTER_UP
            # Predicts the same as the sample above
            if filter_type == 2:
                for i in range(0, bytes_per_row):
                    current_row[i] = (current_row[i] + prior_row[i]) % 256

            # PNG_FILTER_AVERAGE
            # Predicts the average of the sample to the left and the
            # sample above
            if filter_type == 3:
                for i in range(0, bytes_per_pixel):
                    current_row[i] += int(prior_row[i] / 2)

                for i in range(bytes_per_pixel, bytes_per_row):
                    current_row[i] += (int)(
                        (current_row[i - bytes_per_pixel] + prior_row[i]) / 2
                    )
                    current_row[i] %= 256

            # PNG_FILTER_PAETH
            if filter_type == 4:
                for i in range(0, bytes_per_pixel):
                    current_row[i] += prior_row[i]

                for i in range(bytes_per_pixel, bytes_per_row):
                    a = current_row[i - bytes_per_pixel]
                    b = prior_row[i]
                    c = prior_row[i - bytes_per_pixel]

                    p = a + b - c
                    pa = abs(p - a)
                    pb = abs(p - b)
                    pc = abs(p - c)

                    ret = 0
                    if pa <= pb and pa <= pc:
                        ret = a
                    elif pb <= pc:
                        ret = b
                    else:
                        ret = c

                    current_row[i] = (current_row[i] + ret) % 256

            # write current row
            for i in range(0, len(current_row)):
                bytes_after_predictor.append(current_row[i])

            # Swap curr and prior
            prior_row = copy.deepcopy(current_row)

        # return
        return bytes([(int(x) % 256) for x in bytes_after_predictor])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of ReadBaseTransformer is responsible for reading Decimal objects
"""
import io
import typing
from typing import Any, Optional, Union

from borb.io.read.transformer import ReadTransformerState, Transformer
from borb.io.read.types import AnyPDFType, Decimal
from borb.pdf.canvas.event.event_listener import EventListener


class NumberTransformer(Transformer):
    """
    This implementation of ReadBaseTransformer is responsible for reading Decimal objects
    """

    def can_be_transformed(
        self, object: Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO, AnyPDFType]
    ) -> bool:
        """
        This function returns True if the object to be transformed is a Decimal object
        """
        return isinstance(object, Decimal)

    def transform(
        self,
        object_to_transform: Union[io.BufferedIOBase, io.RawIOBase, AnyPDFType],
        parent_object: Any,
        context: Optional[ReadTransformerState] = None,
        event_listeners: typing.List[EventListener] = [],
    ) -> Any:
        """
        This function reads a Decimal from a byte stream
        """
        # fmt: off
        assert isinstance(object_to_transform, Decimal), "object_to_transform must be of type Decimal"
        # fmt: on
        return Decimal(object_to_transform).set_parent(parent_object)  # type: ignore [attr-defined]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This implementation of EventListener extracts all text from a PDF Document,
    substituting composite glyphs for their simpler representations
"""
import typing

from borb.toolkit.text.simple_text_extraction import SimpleTextExtraction


class SimpleNonLigatureTextExtraction(SimpleTextExtraction):
    """
    This implementation of EventListener extracts all text from a PDF Document,
    substituting composite glyphs for their simpler representations
    """

    def __init__(self):
        super(SimpleNonLigatureTextExtraction, self).__init__()
        self._ligatures_to_replace: typing.Dict[str, str] = {
            "Ꜳ": "AA",
            "ꜳ": "aa",
            "Æ": "AE",
            "æ": "ae",
            "ꬱ": "aə",
            "Ꜵ": "AO",
            "ꜵ": "ao",
            "Ꜷ": "AU",
            "ꜷ": "au",
            "Ꜹ": "AV",
            "ꜹ": "av",
            "Ꜻ": "AV",
            "ꜻ": "av",
            "Ꜽ": "AY",
            "ꜽ": "ay",
            "🙰": "et",
            "ꭁ": "əø",
            "ﬀ": "ff",
            "ﬃ": "ffi",
            "ﬄ": "ffl",
            "ﬁ": "fi",
            "ﬂ": "fl",
            "℔": "lb",
            "Ỻ": "IL",
            "ỻ": "ll",
            "Œ": "OE",
            "œ": "oe",
            "Ꝏ": "OO",
            "ꝏ": "oo",
            "ꭢ": "ɔe",
            "ſs": "ẞ",
            "ſz": "ß",
            "ﬆ": "st",
            "ﬅ": "ſt",
            "Ꜩ": "TZ",
            "ꜩ": "tz",
            "ᵫ": "ue",
            "ꭣ": "uo",
            "W": "VV",
            "w": "vv",
            "Ꝡ": "VY",
            "ꝡ": "vy",
        }

    def get_text_for_page(self, page_nr: int) -> str:
        """
        This function returns all text on a given page
        """
        text: str = ""
        if page_nr in self._text_per_page:
            text = self._text_per_page[page_nr]
            while any([k in text for k, v in self._ligatures_to_replace.items()]):
                for k, v in self._ligatures_to_replace.items():
                    text = text.replace(k, v)
        return text



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
In mathematics, a matrix (plural matrices) is a rectangular array or table of numbers,
symbols, or expressions, arranged in rows and columns.
"""
import typing
from decimal import Decimal
from typing import List


class Matrix:
    """
    In mathematics, a matrix (plural matrices) is a rectangular array or table of numbers, symbols, or expressions, arranged in rows and columns.
    Provided that they have the same size (each matrix has the same number of rows and the same number of columns as the other),
    two matrices can be added or subtracted element by element (see conformable matrix).

    The rule for matrix multiplication, however, is that two matrices can be multiplied only when the number of columns
    in the first equals the number of rows in the second (i.e., the inner dimensions are the same, n for an (m×n)-matrix times an (n×p)-matrix,
    resulting in an (m×p)-matrix). There is no product the other way round—a first hint that matrix multiplication is not commutative.
    Any matrix can be multiplied element-wise by a scalar from its associated field.
    """

    def __init__(self):
        """
        Initialize a new Matrix
        """
        self.mtx: typing.List[typing.List[Decimal]] = [[], [], []]

    @staticmethod
    def identity_matrix() -> "Matrix":
        """
        The identity matrix In of size n is the n-by-n matrix in which all the elements on the main diagonal are equal to 1 and all other elements are equal to 0.
        """
        m = Matrix()
        m.mtx = [
            [Decimal(1), Decimal(0), Decimal(0)],
            [Decimal(0), Decimal(1), Decimal(0)],
            [Decimal(0), Decimal(0), Decimal(1)],
        ]
        return m

    @staticmethod
    def matrix_from_six_values(
        a: Decimal, b: Decimal, c: Decimal, d: Decimal, e: Decimal, f: Decimal
    ):
        """
        This method returns the matrix [[a, b, 0], [c, d, 0], [e, f, 1]]
        """
        m = Matrix()
        m.mtx = [[a, b, Decimal(0)], [c, d, Decimal(0)], [e, f, Decimal(1)]]
        return m

    def mul(self, y: "Matrix") -> "Matrix":
        """
        This function multiplies this Matrix with another Matrix,
        returning the result
        """
        m_vals = [
            [Decimal(0), Decimal(0), Decimal(0)],
            [Decimal(0), Decimal(0), Decimal(0)],
            [Decimal(0), Decimal(0), Decimal(0)],
        ]
        for i in range(3):
            for j in range(3):
                for k in range(3):
                    m_vals[i][j] += self.mtx[i][k] * y.mtx[k][j]
        m = Matrix()
        m.mtx = m_vals
        return m

    def cross(self, x: Decimal, y: Decimal, z: Decimal):
        """
        This method calculates the dot-product of this Matrix
        with an input vector (represented by 3 input Decimal objects)
        and returns the result
        """
        x2 = x * self[0][0] + y * self[1][0] + z * self[2][0]
        y2 = x * self[0][1] + y * self[1][1] + z * self[2][1]
        z2 = x * self[0][2] + y * self[1][2] + z * self[2][2]
        return x2, y2, z2

    def __getitem__(self, item) -> List[Decimal]:
        return self.mtx[item]

    def __str__(self):
        return "[[%f %f %f]\n [%f %f %f]\n [%f %f %f]]" % (
            self.mtx[0][0],
            self.mtx[0][1],
            self.mtx[0][2],
            self.mtx[1][0],
            self.mtx[1][1],
            self.mtx[1][2],
            self.mtx[2][0],
            self.mtx[2][1],
            self.mtx[2][2],
        )

    def determinant(self) -> Decimal:
        """
        In linear algebra, the determinant is a scalar value that can be computed from the elements of a square matrix
        and encodes certain properties of the linear transformation described by the matrix.
        The determinant of a matrix A is denoted det(A), det A, or |A|.
        Geometrically, it can be viewed as the volume scaling factor of the linear transformation described by the matrix.
        This is also the signed volume of the n-dimensional parallelepiped spanned by the column or row vectors of the matrix.
        The determinant is positive or negative according to whether the linear transformation preserves or reverses the orientation of a real vector space.
        """
        return (
            self.mtx[0][0]
            * (self.mtx[1][1] * self.mtx[2][2] - self.mtx[1][2] * self.mtx[2][1])
            - self.mtx[0][1]
            * (self.mtx[1][0] * self.mtx[2][2] - self.mtx[1][2] * self.mtx[2][0])
            + self.mtx[0][2]
            * (self.mtx[1][0] * self.mtx[2][1] - self.mtx[1][1] * self.mtx[2][0])
        )

    def __deepcopy__(self, memodict={}):
        m = Matrix()
        m.mtx = [
            [self.mtx[0][0], self.mtx[0][1], self.mtx[0][2]],
            [self.mtx[1][0], self.mtx[1][1], self.mtx[1][2]],
            [self.mtx[2][0], self.mtx[2][1], self.mtx[2][2]],
        ]
        return m



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of LayoutElement represents an ordered (that is to say numbered) list
"""
import typing
from decimal import Decimal

from borb.pdf.canvas.color.color import Color, HexColor, X11Color
from borb.pdf.canvas.layout.layout_element import LayoutElement, Alignment
from borb.pdf.canvas.layout.list.list import List
from borb.pdf.canvas.layout.text.chunk_of_text import ChunkOfText


class OrderedList(List):
    """
    This implementation of LayoutElement represents an ordered (that is to say numbered) list
    """

    def __init__(
        self,
        background_color: typing.Optional[Color] = None,
        border_bottom: bool = False,
        border_color: Color = HexColor("000000"),
        border_left: bool = False,
        border_radius_bottom_left: Decimal = Decimal(0),
        border_radius_bottom_right: Decimal = Decimal(0),
        border_radius_top_left: Decimal = Decimal(0),
        border_radius_top_right: Decimal = Decimal(0),
        border_right: bool = False,
        border_top: bool = False,
        border_width: Decimal = Decimal(1),
        horizontal_alignment: Alignment = Alignment.LEFT,
        margin_bottom: Decimal = None,
        margin_left: Decimal = None,
        margin_right: Decimal = None,
        margin_top: Decimal = None,
        padding_bottom: Decimal = Decimal(0),
        padding_left: Decimal = Decimal(0),
        padding_right: Decimal = Decimal(0),
        padding_top: Decimal = Decimal(0),
        vertical_alignment: Alignment = Alignment.TOP,
    ):
        super(OrderedList, self).__init__(
            border_top=border_top,
            border_right=border_right,
            border_bottom=border_bottom,
            border_left=border_left,
            border_radius_top_left=border_radius_top_left,
            border_radius_top_right=border_radius_top_right,
            border_radius_bottom_right=border_radius_bottom_right,
            border_radius_bottom_left=border_radius_bottom_left,
            border_color=border_color,
            border_width=border_width,
            padding_top=padding_top,
            padding_right=padding_right,
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            margin_left=margin_left,
            background_color=background_color,
            horizontal_alignment=horizontal_alignment,
            vertical_alignment=vertical_alignment,
        )

    def _get_bullet_layout_element(
        self, item_index: int, item: LayoutElement
    ) -> LayoutElement:
        return ChunkOfText(
            text=str(item_index + 1) + ".",
            font_size=self.get_font_size(),
            font_color=X11Color("Black"),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    In computer science and visualization, a canvas is a container that holds various drawing elements
    (lines, shapes, text, frames containing other elements, etc.).
    It takes its name from the canvas used in visual arts.
"""

from borb.io.read.types import Dictionary
from borb.pdf.canvas.canvas_graphics_state import CanvasGraphicsState


class Canvas(Dictionary):
    """
    In computer science and visualization, a canvas is a container that holds various drawing elements
    (lines, shapes, text, frames containing other elements, etc.).
    It takes its name from the canvas used in visual arts.
    """

    def __init__(self):
        super(Canvas, self).__init__()
        # compatibility mode
        self.in_compatibility_section = False
        # set initial graphics state
        self.graphics_state = CanvasGraphicsState()
        # canvas tag hierarchy is (oddly enough) not considered to be part of the graphics state
        self.marked_content_stack = []
        # set graphics state stack
        self.graphics_state_stack = []



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This class converts HTML to PDF.
"""
import copy
import logging
import typing
import xml.etree.ElementTree as ET

from lxml.etree import HTMLParser, _Comment  # type: ignore [import]

from borb.io.read.types import Decimal, Name, String, Dictionary
from borb.pdf import Image
from borb.pdf.canvas.color.color import Color, HexColor
from borb.pdf.canvas.font.font import Font
from borb.pdf.canvas.font.simple_font.font_type_1 import StandardType1Font
from borb.pdf.canvas.layout.emoji.emoji import Emoji, Emojis
from borb.pdf.canvas.layout.page_layout.multi_column_layout import SingleColumnLayout
from borb.pdf.canvas.layout.list.unordered_list import UnorderedList
from borb.pdf.canvas.layout.list.ordered_list import OrderedList
from borb.pdf.canvas.layout.table.flexible_column_width_table import (
    FlexibleColumnWidthTable,
)
from borb.pdf.canvas.layout.table.table import Table
from borb.pdf.canvas.layout.horizontal_rule import HorizontalRule
from borb.pdf.canvas.layout.layout_element import LayoutElement
from borb.pdf.canvas.layout.page_layout.page_layout import PageLayout
from borb.pdf.canvas.layout.text.chunk_of_text import ChunkOfText
from borb.pdf.canvas.layout.text.heterogeneous_paragraph import LineBreakChunk
from borb.pdf.document.document import Document
from borb.pdf.page.page import Page
from borb.pdf.xref.plaintext_xref import PlainTextXREF
from borb.pdf.canvas.layout.page_layout.block_flow import BlockFlow
from borb.pdf.canvas.layout.page_layout.inline_flow import InlineFlow

logger = logging.getLogger(__name__)


class HTMLToPDF:
    """
    This class converts HTML to PDF
    """

    class Context:
        """
        This class represents the context in which an (HTML) Element is
        being transformed into its LayoutElement representation. This context
        keeps track of whether an Element should be rendered bold, italic, or both,
        the font_size and the background_color.
        """

        def __init__(self):
            # font properties
            self._fallback_fonts_regular: typing.List[Font] = [
                StandardType1Font("Helvetica")
            ]
            self._fallback_fonts_italic: typing.List[Font] = [
                StandardType1Font("Helvetica-Oblique")
            ]
            self._fallback_fonts_bold: typing.List[Font] = [
                StandardType1Font("Helvetica-Bold")
            ]
            self._fallback_fonts_bold_italic: typing.List[Font] = [
                StandardType1Font("Helvetica-Bold-Oblique")
            ]
            self._fallback_fonts_monospaced: typing.List[Font] = [
                StandardType1Font("Courier")
            ]
            self._is_bold: bool = False
            self._is_italic: bool = False
            self._is_monospaced: bool = False
            self._default_font_size: Decimal = Decimal(12)
            self._font_size: Decimal = self._default_font_size
            self._font_color: typing.Optional[Color] = HexColor("000000")
            # style properties
            self._background_color: typing.Optional[Color] = None
            # linking
            self._document: typing.Optional[Document] = None

    @staticmethod
    def _process_element(e: ET.Element, c: Context) -> LayoutElement:
        ELEMENT_CREATOR_METHODS: typing.Dict[
            str, typing.Callable[[ET.Element, HTMLToPDF.Context], LayoutElement]
        ] = {
            "a": HTMLToPDF._process_a_element,
            "abbr": None,
            "acronym": None,
            "address": None,  # TODO
            "article": HTMLToPDF._process_article_element,
            "aside": None,
            "b": HTMLToPDF._process_b_element,
            "bdo": None,
            "big": None,  # TODO
            "blockquote": HTMLToPDF._process_blockquote_element,
            "body": HTMLToPDF._process_body_element,
            "br": HTMLToPDF._process_br_element,
            "button": None,
            "canvas": None,
            "cite": None,  # TODO
            "code": HTMLToPDF._process_code_element,
            "dd": None,
            "dfn": None,
            "div": HTMLToPDF._process_div_element,
            "dl": None,
            "dt": None,
            "em": HTMLToPDF._process_em_element,
            "fieldset": None,
            "figcaption": None,  # TODO
            "figure": HTMLToPDF._process_figure_element,
            "footer": None,  # TODO
            "form": None,
            "h1": HTMLToPDF._process_hx_element,
            "h2": HTMLToPDF._process_hx_element,
            "h3": HTMLToPDF._process_hx_element,
            "h4": HTMLToPDF._process_hx_element,
            "h5": HTMLToPDF._process_hx_element,
            "h6": HTMLToPDF._process_hx_element,
            "head": HTMLToPDF._process_head_element,
            "header": None,  # TODO
            "hr": HTMLToPDF._process_hr_element,
            "html": HTMLToPDF._process_html_element,
            "i": HTMLToPDF._process_i_element,
            "img": HTMLToPDF._process_img_element,
            "kbd": None,
            "li": HTMLToPDF._process_li_element,
            "main": HTMLToPDF._process_main_element,
            "map": None,  # TODO
            "mark": HTMLToPDF._process_mark_element,
            "meta": HTMLToPDF._process_meta_element,
            "nav": None,
            "noscript": HTMLToPDF._process_noscript_element,
            "ol": HTMLToPDF._process_ol_element,
            "p": HTMLToPDF._process_p_element,
            "pre": None,  # TODO
            "q": HTMLToPDF._process_q_element,
            "samp": HTMLToPDF._process_samp_element,
            "section": HTMLToPDF._process_section_element,
            "span": HTMLToPDF._process_inline_element,
            "strong": HTMLToPDF._process_strong_element,
            "table": HTMLToPDF._process_table_element,
            "title": HTMLToPDF._process_title_element,
            "tfoot": None,
            "ul": HTMLToPDF._process_ul_element,
            "video": None,  # TODO
        }
        if ELEMENT_CREATOR_METHODS.get(e.tag, None) is None:
            logger.warning("<%s> unsupported" % e.tag)
            return None
        return ELEMENT_CREATOR_METHODS[e.tag](e, c)

    @staticmethod
    def _process_a_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "a"
        e.tag = "span"

        prev_tail: str = e.tail
        e.tail = ""
        prev_font_color = c._font_color
        c._font_color = HexColor("#0645ad")
        out_value_001: LayoutElement = HTMLToPDF._process_element(e, c)
        c._font_color = prev_font_color
        e.tag = "a"

        # fake span to hold tail
        tmp: ET.Element = ET.Element("span")
        tmp.text = prev_tail
        out_value_002: LayoutElement = HTMLToPDF._process_element(tmp, c)
        e.tail = prev_tail

        # return
        return InlineFlow().extend([out_value_001, out_value_002])

    @staticmethod
    def _process_article_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "article"
        return HTMLToPDF._process_block_element(e, c)

    @staticmethod
    def _process_b_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "b"
        e.tag = "strong"
        out_value: LayoutElement = HTMLToPDF._process_element(e, c)
        e.tag = "b"
        return out_value

    @staticmethod
    def _process_blockquote_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "blockquote"
        prev_background_color: typing.Optional[Color] = c._background_color
        c._background_color = HexColor("F5F5F5")
        blockquote_element: LayoutElement = HTMLToPDF._process_block_element(e, c)
        blockquote_element._background_color = c._background_color
        blockquote_element._padding_left = c._font_size * 2
        blockquote_element._border_left = True
        blockquote_element._border_width = Decimal(2)
        blockquote_element._border_color = HexColor("DCDCDC")
        c._background_color = prev_background_color
        return blockquote_element

    @staticmethod
    def _process_body_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "body"
        return HTMLToPDF._process_block_element(e, c)

    @staticmethod
    def _process_br_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "br"
        e.tag = "span"
        out_value: LayoutElement = HTMLToPDF._process_element(e, c)
        e.tag = "br"
        return InlineFlow().extend([LineBreakChunk(), out_value])

    @staticmethod
    def _process_code_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "code"
        e.tag = "span"

        prev_tail: str = e.tail
        e.tail = ""
        prev_background_color = c._background_color
        prev_is_monospaced: bool = c._is_monospaced
        c._background_color = HexColor("#f5f5f5")
        c._is_monospaced = True
        out_value_001: LayoutElement = HTMLToPDF._process_element(e, c)
        c._background_color = prev_background_color
        c._is_monospaced = prev_is_monospaced
        e.tag = "code"

        # fake span to hold tail
        tmp: ET.Element = ET.Element("span")
        tmp.text = prev_tail
        out_value_002: LayoutElement = HTMLToPDF._process_element(tmp, c)
        e.tail = prev_tail

        # return
        return InlineFlow().extend([out_value_001, out_value_002])

    @staticmethod
    def _process_div_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "div"
        return HTMLToPDF._process_element(e, c)

    @staticmethod
    def _process_em_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "em"
        e.tag = "span"

        prev_tail: str = e.tail
        e.tail = ""
        prev_is_italic = c._is_italic
        c._is_italic = True
        out_value_001: LayoutElement = HTMLToPDF._process_element(e, c)
        c._is_italic = prev_is_italic
        e.tag = "em"

        # fake span to hold tail
        tmp: ET.Element = ET.Element("span")
        tmp.text = prev_tail
        out_value_002: LayoutElement = HTMLToPDF._process_element(tmp, c)
        e.tail = prev_tail

        # return
        return InlineFlow().extend([out_value_001, out_value_002])

    @staticmethod
    def _process_figure_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "figure"
        figcaption: typing.Optional[ET.Element] = next(
            iter([x for x in e if x.tag == "figcaption"]), None
        )
        img: typing.Optional[ET.Element] = next(
            iter([x for x in e if x.tag == "img"]), None
        )
        number_of_rows: int = 0
        if figcaption is not None:
            number_of_rows += 1
        if img is not None:
            number_of_rows += 1
        table: Table = FlexibleColumnWidthTable(
            number_of_columns=1, number_of_rows=number_of_rows
        )
        if img is not None:
            table.add(HTMLToPDF._process_element(img, c))
        if figcaption is not None:
            table.add(HTMLToPDF._process_element(figcaption, c))
        table.set_padding_on_all_cells(Decimal(2), Decimal(2), Decimal(2), Decimal(2))
        table.no_borders()
        return table

    @staticmethod
    def _process_hx_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag in ["h1", "h2", "h3", "h4", "h5", "h6"]
        FONT_SIZE_MODIFIER: typing.Dict[str, Decimal] = {
            "h1": Decimal(2),
            "h2": Decimal(1.5),
            "h3": Decimal(1.17),
            "h4": Decimal(1),
            "h5": Decimal(0.83),
            "h6": Decimal(0.67),
        }
        # modify Context
        prev_font_size: Decimal = c._font_size
        prev_is_bold: bool = c._is_bold
        c._font_size = c._default_font_size * FONT_SIZE_MODIFIER[e.tag]
        c._is_bold = True

        # recursion
        header_layout_element: LayoutElement = BlockFlow().add(
            HTMLToPDF._process_inline_element(e, c)
        )
        header_layout_element._padding_bottom = c._font_size
        header_layout_element._padding_top = c._font_size

        # modify Context
        c._font_size = prev_font_size
        c._is_bold = prev_is_bold

        # return
        return header_layout_element

    @staticmethod
    def _process_head_element(e: ET.Element, c: Context) -> None:
        assert e.tag == "head"
        # title
        if "title" in [x.tag for x in e]:
            HTMLToPDF._process_element([x for x in e if x.tag == "title"][0], c)
        # meta
        for meta_element in [x for x in e if x.tag == "meta"]:
            HTMLToPDF._process_element(meta_element, c)

    @staticmethod
    def _process_hr_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "hr"

        # process tail
        tmp: ET.Element = ET.Element("span")
        tmp.text = e.tail
        out_value_002: LayoutElement = HTMLToPDF._process_element(tmp, c)

        # return
        return BlockFlow().extend([HorizontalRule(), out_value_002])

    @staticmethod
    def _process_html_element(e: ET.Element, c: Context) -> LayoutElement:
        # head
        if "head" in [x.tag for x in e]:
            HTMLToPDF._process_element([x for x in e if x.tag == "head"][0], c)
        # body
        if "body" in [x.tag for x in e]:
            return HTMLToPDF._process_element([x for x in e if x.tag == "body"][0], c)
        # default
        return BlockFlow()

    @staticmethod
    def _process_i_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "i"
        e.tag = "em"
        out_value: LayoutElement = HTMLToPDF._process_element(e, c)
        e.tag = "i"
        return out_value

    @staticmethod
    def _process_img_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "img"
        src: str = e.attrib["src"]
        w: typing.Optional[Decimal] = (
            Decimal(e.attrib["width"]) if "width" in e.attrib else None
        )
        h: typing.Optional[Decimal] = (
            Decimal(e.attrib["height"]) if "height" in e.attrib else None
        )
        out_value_001: LayoutElement = Image(src, width=w, height=h)

        # process tail
        tmp: ET.Element = ET.Element("span")
        tmp.text = e.tail
        out_value_002: LayoutElement = HTMLToPDF._process_element(tmp, c)

        # return
        return InlineFlow().extend([out_value_001, out_value_002])

    @staticmethod
    def _process_li_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "li"
        return HTMLToPDF._process_block_element(e, c)

    @staticmethod
    def _process_main_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "main"
        return HTMLToPDF._process_block_element(e, c)

    @staticmethod
    def _process_mark_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "mark"
        e.tag = "span"

        prev_tail: str = e.tail
        e.tail = ""
        prev_background_color = c._background_color
        c._background_color = HexColor("#ffff00")
        out_value_001: LayoutElement = HTMLToPDF._process_element(e, c)
        c._background_color = prev_background_color
        e.tag = "mark"

        # fake span to hold tail
        tmp: ET.Element = ET.Element("span")
        tmp.text = prev_tail
        out_value_002: LayoutElement = HTMLToPDF._process_element(tmp, c)
        e.tail = prev_tail

        # return
        return InlineFlow().extend([out_value_001, out_value_002])

    @staticmethod
    def _process_meta_element(e: ET.Element, c: Context) -> None:
        assert e.tag == "meta"
        document: typing.Optional[Document] = c._document
        if document is None:
            return
        name: str = e.attrib.get("name", "")
        content: str = e.attrib.get("content", "")
        if "XRef" not in document:
            document[Name("XRef")] = PlainTextXREF()
        if "Trailer" not in document["XRef"]:
            document["XRef"][Name("Trailer")] = Dictionary()
        if "Info" not in document["XRef"]["Trailer"]:
            document["XRef"]["Trailer"][Name("Info")] = Dictionary()
        if name == "keywords":
            document["XRef"]["Trailer"]["Info"][Name("Keywords")] = String(content)
        if name == "description":
            document["XRef"]["Trailer"]["Info"][Name("Subject")] = String(content)
        if name == "author":
            document["XRef"]["Trailer"]["Info"][Name("Author")] = String(content)

    @staticmethod
    def _process_noscript_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "noscript"
        e.tag = "p"
        return_value: LayoutElement = HTMLToPDF._process_element(e, c)
        e.tag = "noscript"
        return return_value

    @staticmethod
    def _process_ol_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "ol"
        ol_layout_element: OrderedList = OrderedList()
        for child_element in e:
            if child_element.tag == ET.Comment:
                continue
            ol_layout_element.add(HTMLToPDF._process_element(child_element, c))
        return BlockFlow().add(ol_layout_element)

    @staticmethod
    def _process_p_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "p"
        return BlockFlow().add(HTMLToPDF._process_inline_element(e, c))

    @staticmethod
    def _process_pre_element(e: ET.Element, c: Context) -> LayoutElement:
        return None

    @staticmethod
    def _process_q_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "q"
        e.tag = "span"
        out_value: LayoutElement = HTMLToPDF._process_element(e, c)
        e.tag = "q"
        return InlineFlow().extend(
            [
                HTMLToPDF._build_chunk_of_text('"', c),
                out_value,
                HTMLToPDF._build_chunk_of_text('"', c),
            ]
        )

    @staticmethod
    def _process_samp_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "samp"
        e.tag = "span"
        prev_is_monospaced: bool = c._is_monospaced
        c._is_monospaced = True
        out_value: LayoutElement = HTMLToPDF._process_element(e, c)
        c._is_monospaced = prev_is_monospaced
        e.tag = "samp"
        return out_value

    @staticmethod
    def _process_section_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "section"
        e.tag = "div"
        out_value: LayoutElement = HTMLToPDF._process_element(e, c)
        e.tag = "section"
        return out_value

    @staticmethod
    def _process_strong_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "strong"
        e.tag = "span"

        prev_tail: str = e.tail
        e.tail = ""
        prev_is_bold = c._is_bold
        c._is_bold = True
        out_value_001: LayoutElement = HTMLToPDF._process_element(e, c)
        c._is_bold = prev_is_bold
        e.tag = "strong"

        # fake span to hold tail
        tmp: ET.Element = ET.Element("span")
        tmp.text = prev_tail
        out_value_002: LayoutElement = HTMLToPDF._process_element(tmp, c)
        e.tail = prev_tail

        # return
        return InlineFlow().extend([out_value_001, out_value_002])

    @staticmethod
    def _process_table_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "table"
        if "tbody" in [x.tag for x in e]:
            return HTMLToPDF._process_table_element_001(e, c)
        else:
            return HTMLToPDF._process_table_element_002(e, c)

    @staticmethod
    def _process_table_element_001(e: ET.Element, c: Context) -> LayoutElement:
        # separate thead / tbody
        thead: typing.Optional[ET.Element] = next(
            iter([x for x in e if x.tag == "thead"]), None
        )
        tbody: ET.Element = next(iter([x for x in e if x.tag == "tbody"]))

        # count rows / cols
        nrows_000: int = len([x for x in tbody if x.tag == "tr"]) + (
            1 if thead is not None else 0
        )
        ncols_000: int = len(
            [y for y in [x for x in tbody if x.tag == "tr"][0] if y.tag == "td"]
        )

        # create Table
        table: Table = FlexibleColumnWidthTable(
            number_of_rows=nrows_000, number_of_columns=ncols_000
        )
        if thead is not None:
            for tr in [x for x in thead if x.tag == "tr"]:
                for th in [y for y in tr if y.tag == "th"]:
                    table.add(HTMLToPDF._process_inline_element(th, c))
        for tr in [x for x in tbody if x.tag == "tr"]:
            for td in [y for y in tr if y.tag == "td"]:
                table.add(HTMLToPDF._process_inline_element(td, c))

        # add Table to body
        table.set_padding_on_all_cells(Decimal(2), Decimal(2), Decimal(2), Decimal(2))
        table._margin_top = c._default_font_size
        table._margin_bottom = c._default_font_size
        return table

    @staticmethod
    def _process_table_element_002(e: ET.Element, c: Context) -> LayoutElement:

        # count rows / cols
        nrows_001: int = len([x for x in e if x.tag == "tr"])
        ncols_001: int = len(
            [y for y in [x for x in e if x.tag == "tr"][0] if y.tag == "td"]
        )

        # create Table
        table: Table = FlexibleColumnWidthTable(
            number_of_rows=nrows_001, number_of_columns=ncols_001
        )
        for tr in [x for x in e if x.tag == "tr"]:
            for td in [y for y in tr if y.tag == "td"]:
                table.add(HTMLToPDF._process_inline_element(td, c))

        # add Table to body
        table.set_padding_on_all_cells(Decimal(2), Decimal(2), Decimal(2), Decimal(2))
        table._margin_top = c._default_font_size
        table._margin_bottom = c._default_font_size
        return table

    @staticmethod
    def _process_title_element(e: ET.Element, c: Context) -> None:
        assert e.tag == "title"
        document: typing.Optional[Document] = c._document
        if document is None:
            return
        if "XRef" not in document:
            document[Name("XRef")] = PlainTextXREF()
        if "Trailer" not in document["XRef"]:
            document["XRef"][Name("Trailer")] = Dictionary()
        if "Info" not in document["XRef"]["Trailer"]:
            document["XRef"]["Trailer"][Name("Info")] = Dictionary()
        document["XRef"]["Trailer"]["Info"][Name("Title")] = String(e.text)

    @staticmethod
    def _process_ul_element(e: ET.Element, c: Context) -> LayoutElement:
        assert e.tag == "ul"
        ul_layout_element: UnorderedList = UnorderedList()
        for child_element in e:
            if child_element.tag == ET.Comment:
                continue
            ul_layout_element.add(HTMLToPDF._process_element(child_element, c))
        return ul_layout_element

    #
    # LOW-LEVEL METHODS
    #

    @staticmethod
    def _process_block_element(e: ET.Element, c: Context) -> LayoutElement:
        block_layout_element: BlockFlow = BlockFlow()

        # add text
        # create artificial <span> element
        # then pass that element through HTMLToPDF._process_element
        if e.text is not None and len(e.text) > 0:
            tmp_span: ET.Element = ET.Element("span")
            tmp_span.text = e.text
            block_layout_element.add(HTMLToPDF._process_element(tmp_span, c))

        # process children
        for child_element in e:
            if isinstance(child_element, _Comment):
                continue
            block_layout_element.add(HTMLToPDF._process_element(child_element, c))

        # add tail
        # create artificial <span> element
        # then pass that element through HTMLToPDF._process_element
        if e.tail is not None and len(e.tail) > 0:
            tmp_span: ET.Element = ET.Element("span")
            tmp_span.text = e.tail
            block_layout_element.add(HTMLToPDF._process_element(tmp_span, c))

        # return
        return block_layout_element

    @staticmethod
    def _process_inline_element(e: ET.Element, c: Context) -> LayoutElement:

        # <span class="emoji"></span>
        chunks: typing.List[LayoutElement] = []
        if e.tag == "span" and (
            "EMOJI" in e.attrib.get("class", "").upper().split(" ")
        ):
            emoji: str = [
                x
                for x in e.attrib.get("class", "").upper().split(" ")
                if x.startswith("EMOJI_")
            ][0][6:]
            chunks.append(Emojis[emoji].value)
            e.text = ""

        # text
        if e.text is not None and len(e.text) > 0:
            ws: typing.List[str] = [x for x in e.text.split(" ")]
            ws = [x + " " for x in ws[0 : len(ws) - 1]] + [ws[-1]]
            chunks.extend([HTMLToPDF._build_chunk_of_text(w, c) for w in ws])

        # children
        for child_element in e:
            if isinstance(child_element, _Comment):
                continue
            chunks.append(HTMLToPDF._process_element(child_element, c))

        # tail
        if e.tail is not None and len(e.tail) > 0:
            ws: typing.List[str] = [x for x in e.tail.split(" ")]
            ws = [x + " " for x in ws[0 : len(ws) - 1]] + [ws[-1]]
            chunks.extend([HTMLToPDF._build_chunk_of_text(w, c) for w in ws])

        # empty
        if len(chunks) == 0:
            chunks.append(HTMLToPDF._build_chunk_of_text(" ", c))

        # return
        return InlineFlow().extend(chunks)

    @staticmethod
    def _build_chunk_of_text(s: str, c: Context) -> LayoutElement:
        fonts_to_try: typing.List[Font] = []
        if c._is_bold and c._is_italic:
            fonts_to_try = c._fallback_fonts_bold_italic
        elif c._is_bold:
            fonts_to_try = c._fallback_fonts_bold
        elif c._is_italic:
            fonts_to_try = c._fallback_fonts_italic
        else:
            fonts_to_try = c._fallback_fonts_regular
        if c._is_monospaced:
            fonts_to_try = c._fallback_fonts_monospaced

        # try
        for font_to_try in fonts_to_try:
            try:
                c: ChunkOfText = ChunkOfText(
                    s,
                    font=font_to_try,
                    font_size=c._font_size,
                    font_color=c._font_color,
                    background_color=c._background_color,
                )
                c._write_text_bytes()
                return c
            except:
                continue

        # default
        return ChunkOfText(
            "□",
            font="Helvetica",
            font_size=c._font_size,
            font_color=c._font_color,
            background_color=c._background_color,
        )

    #
    # public methods
    #

    @staticmethod
    def convert_html_to_layout_element(
        html: typing.Union[str, ET.Element],
        fallback_fonts_regular: typing.List[Font] = [StandardType1Font("Helvetica")],
        fallback_fonts_bold: typing.List[Font] = [StandardType1Font("Helvetica-Bold")],
        fallback_fonts_italic: typing.List[Font] = [
            StandardType1Font("Helvetica-Oblique")
        ],
        fallback_fonts_bold_italic: typing.List[Font] = [
            StandardType1Font("Helvetica-Bold-Oblique")
        ],
        fallback_fonts_monospaced: typing.List[Font] = [StandardType1Font("Courier")],
    ) -> LayoutElement:
        """
        This function converts a html str to a LayoutElement
        :param html:                        the html str (or ET.Element) to be converted
        :param fallback_fonts_regular:      fallback (regular) fonts to try when the default font is unable to render a character
        :param fallback_fonts_bold:         fallback (bold) fonts to try when the default font is unable to render a character
        :param fallback_fonts_italic:       fallback (italic) fonts to try when the default font is unable to render a character
        :param fallback_fonts_bold_italic:  fallback (bold, italic) fonts to try when the default font is unable to render a character
        :return:
        """

        # convert str to ET.Element
        root_element: typing.Optional[ET.Element] = None
        if isinstance(html, str):
            root_element = ET.fromstring(html, HTMLParser())
        else:
            root_element = html
        assert root_element is not None

        # build context
        c: HTMLToPDF.Context = HTMLToPDF.Context()
        c._fallback_fonts_regular = fallback_fonts_regular
        c._fallback_fonts_bold = fallback_fonts_bold
        c._fallback_fonts_italic = fallback_fonts_italic
        c._fallback_fonts_bold_italic = fallback_fonts_bold_italic
        c._fallback_fonts_monospaced = fallback_fonts_monospaced

        # return
        return HTMLToPDF._process_html_element(root_element, c)

    @staticmethod
    def convert_html_to_pdf(
        html: typing.Union[str, ET.Element],
        fallback_fonts_regular: typing.List[Font] = [StandardType1Font("Helvetica")],
        fallback_fonts_bold: typing.List[Font] = [StandardType1Font("Helvetica-Bold")],
        fallback_fonts_italic: typing.List[Font] = [
            StandardType1Font("Helvetica-Oblique")
        ],
        fallback_fonts_bold_italic: typing.List[Font] = [
            StandardType1Font("Helvetica-Bold-Oblique")
        ],
        fallback_fonts_monospaced: typing.List[Font] = [StandardType1Font("Courier")],
    ) -> Document:
        """
        This function converts HTML to PDF
        """

        # Document
        doc: Document = Document()

        # Page
        page: Page = Page()
        doc.add_page(page)

        # PageLayout
        layout: PageLayout = SingleColumnLayout(page)

        # parse HTML
        root_element: typing.Optional[ET.Element] = None
        if isinstance(html, str):
            root_element = ET.fromstring(html, HTMLParser())
        else:
            root_element = html
        assert root_element is not None

        # build Context
        c: HTMLToPDF.Context = HTMLToPDF.Context()
        c._document = doc
        c._fallback_fonts_regular = fallback_fonts_regular
        c._fallback_fonts_bold = fallback_fonts_bold
        c._fallback_fonts_italic = fallback_fonts_italic
        c._fallback_fonts_bold_italic = fallback_fonts_bold_italic
        c._fallback_fonts_monospaced = fallback_fonts_monospaced

        # convert HTML to PDF
        layout.add(HTMLToPDF._process_html_element(root_element, c))

        # return
        return doc



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Same as K but used for nonstroking operations.
"""
import typing
from decimal import Decimal

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.color.color import CMYKColor
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class SetCMYKNonStroking(CanvasOperator):
    """
    Same as K but used for nonstroking operations.
    """

    def __init__(self):
        super().__init__("k", 4)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the k operator
        """
        # fmt: off
        assert isinstance(operands[0], Decimal), "Operand 0 of k must be a Decimal"
        assert isinstance(operands[1], Decimal), "Operand 1 of k must be a Decimal"
        assert isinstance(operands[2], Decimal), "Operand 2 of k must be a Decimal"
        assert isinstance(operands[3], Decimal), "Operand 3 of k must be a Decimal"
        # fmt: on
        c = operands[0]
        m = operands[1]
        y = operands[2]
        k = operands[3]
        canvas = canvas_stream_processor.get_canvas()
        canvas.graphics_state.non_stroke_color = CMYKColor(c, m, y, k)
